# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from buildbot.util import lru
from buildbot.util import service


class CacheManager(service.ReconfigurableServiceMixin, service.AsyncService):
    """
    A manager for a collection of caches, each for different types of objects
    and with potentially-overlapping key spaces.

    There is generally only one instance of this class, available at
    C{master.caches}.
    """

    # a cache of length one still has many benefits: it collects objects that
    # remain referenced elsewhere; it collapses simultaneous misses into one
    # miss function; and it will optimize repeated fetches of the same object.
    DEFAULT_CACHE_SIZE = 1

    def __init__(self):
        self.setName('caches')
        self.config = {}
        self._caches = {}

    def get_cache(self, cache_name, miss_fn):
        """
        Get an L{AsyncLRUCache} object with the given name.  If such an object
        does not exist, it will be created.  Since the cache is permanent, this
        method can be called only once, e.g., in C{startService}, and it value
        stored indefinitely.

        @param cache_name: name of the cache (usually the name of the type of
        object it stores)
        @param miss_fn: miss function for the cache; see L{AsyncLRUCache}
        constructor.
        @returns: L{AsyncLRUCache} instance
        """
        try:
            return self._caches[cache_name]
        except KeyError:
            max_size = self.config.get(cache_name, self.DEFAULT_CACHE_SIZE)
            assert max_size >= 1
            c = self._caches[cache_name] = lru.AsyncLRUCache(miss_fn, max_size)
            return c

    def reconfigServiceWithBuildbotConfig(self, new_config):
        self.config = new_config.caches
        for name, cache in self._caches.items():
            cache.set_max_size(new_config.caches.get(name, self.DEFAULT_CACHE_SIZE))

        return super().reconfigServiceWithBuildbotConfig(new_config)

    def get_metrics(self):
        return {
            n: {'hits': c.hits, 'refhits': c.refhits, 'misses': c.misses, 'max_size': c.max_size}
            for n, c in self._caches.items()
        }



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from typing import TYPE_CHECKING

import sqlalchemy as sa
from twisted.internet import defer

from buildbot.db import base
from buildbot.util import identifiers
from buildbot.warnings import warn_deprecated

if TYPE_CHECKING:
    from typing import Any


@dataclass
class BuilderMasterModel:
    builderid: int
    masterid: int

    # For backward compatibility
    def __getitem__(self, key: str):
        warn_deprecated(
            '4.1.0',
            (
                'WorkersConnectorComponent '
                'getWorker, and getWorkers '
                'no longer return Worker as dictionnaries. '
                'Usage of [] accessor is deprecated: please access the member directly'
            ),
        )

        if hasattr(self, key):
            return getattr(self, key)

        raise KeyError(key)


@dataclass
class WorkerModel:
    id: int
    name: str
    workerinfo: dict[str, Any]
    paused: bool = False
    pause_reason: str | None = None
    graceful: bool = False

    configured_on: list[BuilderMasterModel] = field(default_factory=list)
    connected_to: list[int] = field(default_factory=list)

    # For backward compatibility
    def __getitem__(self, key: str):
        warn_deprecated(
            '4.1.0',
            (
                'WorkersConnectorComponent '
                'getWorker, and getWorkers '
                'no longer return Worker as dictionnaries. '
                'Usage of [] accessor is deprecated: please access the member directly'
            ),
        )

        if hasattr(self, key):
            return getattr(self, key)

        raise KeyError(key)


class WorkersConnectorComponent(base.DBConnectorComponent):
    def findWorkerId(self, name):
        tbl = self.db.model.workers
        # callers should verify this and give good user error messages
        assert identifiers.isIdentifier(50, name)
        return self.findSomethingId(
            tbl=tbl,
            whereclause=(tbl.c.name == name),
            insert_values={
                "name": name,
                "info": {},
                "paused": 0,
                "pause_reason": None,
                "graceful": 0,
            },
        )

    def _deleteFromConfiguredWorkers_thd(self, conn, buildermasterids, workerid=None):
        cfg_tbl = self.db.model.configured_workers
        # batch deletes to avoid using too many variables
        for batch in self.doBatch(buildermasterids, 100):
            q = cfg_tbl.delete()
            q = q.where(cfg_tbl.c.buildermasterid.in_(batch))
            if workerid:
                q = q.where(cfg_tbl.c.workerid == workerid)
            conn.execute(q).close()

    # returns a Deferred which returns None
    def deconfigureAllWorkersForMaster(self, masterid):
        def thd(conn):
            # first remove the old configured buildermasterids for this master and worker
            # as sqlalchemy does not support delete with join, we need to do
            # that in 2 queries
            cfg_tbl = self.db.model.configured_workers
            bm_tbl = self.db.model.builder_masters
            j = cfg_tbl
            j = j.outerjoin(bm_tbl)
            q = sa.select(cfg_tbl.c.buildermasterid).select_from(j).distinct()
            q = q.where(bm_tbl.c.masterid == masterid)
            res = conn.execute(q)
            buildermasterids = [row.buildermasterid for row in res]
            res.close()
            self._deleteFromConfiguredWorkers_thd(conn, buildermasterids)

        return self.db.pool.do_with_transaction(thd)

    # returns a Deferred that returns None
    def workerConfigured(self, workerid, masterid, builderids):
        def thd(conn):
            cfg_tbl = self.db.model.configured_workers
            bm_tbl = self.db.model.builder_masters

            # get the buildermasterids that are configured
            if builderids:
                q = sa.select(bm_tbl.c.id).select_from(bm_tbl)
                q = q.where(bm_tbl.c.masterid == masterid)
                q = q.where(bm_tbl.c.builderid.in_(builderids))
                res = conn.execute(q)
                buildermasterids = {row.id for row in res}
                res.close()
            else:
                buildermasterids = set([])

            j = cfg_tbl
            j = j.outerjoin(bm_tbl)
            q = sa.select(cfg_tbl.c.buildermasterid).select_from(j).distinct()
            q = q.where(bm_tbl.c.masterid == masterid)
            q = q.where(cfg_tbl.c.workerid == workerid)
            res = conn.execute(q)
            oldbuildermasterids = {row.buildermasterid for row in res}
            res.close()

            todeletebuildermasterids = oldbuildermasterids - buildermasterids
            toinsertbuildermasterids = buildermasterids - oldbuildermasterids
            self._deleteFromConfiguredWorkers_thd(conn, todeletebuildermasterids, workerid)

            # and insert the new ones
            if toinsertbuildermasterids:
                q = cfg_tbl.insert()
                conn.execute(
                    q,
                    [
                        {'workerid': workerid, 'buildermasterid': buildermasterid}
                        for buildermasterid in toinsertbuildermasterids
                    ],
                ).close()

        return self.db.pool.do_with_transaction(thd)

    @defer.inlineCallbacks
    def getWorker(
        self,
        workerid: int | None = None,
        name: str | None = None,
        masterid: int | None = None,
        builderid: int | None = None,
    ):
        if workerid is None and name is None:
            return None
        workers = yield self.getWorkers(
            _workerid=workerid, _name=name, masterid=masterid, builderid=builderid
        )
        if workers:
            return workers[0]
        return None

    def getWorkers(
        self,
        _workerid: int | None = None,
        _name: str | None = None,
        masterid: int | None = None,
        builderid: int | None = None,
        paused: bool | None = None,
        graceful: bool | None = None,
    ) -> defer.Deferred[list[WorkerModel]]:
        def thd(conn) -> list[WorkerModel]:
            workers_tbl = self.db.model.workers
            conn_tbl = self.db.model.connected_workers
            cfg_tbl = self.db.model.configured_workers
            bm_tbl = self.db.model.builder_masters

            # first, get the worker itself and the configured_on info
            j = workers_tbl
            j = j.outerjoin(cfg_tbl)
            j = j.outerjoin(bm_tbl)
            q = (
                sa.select(
                    workers_tbl.c.id,
                    workers_tbl.c.name,
                    workers_tbl.c.info,
                    workers_tbl.c.paused,
                    workers_tbl.c.pause_reason,
                    workers_tbl.c.graceful,
                    bm_tbl.c.builderid,
                    bm_tbl.c.masterid,
                )
                .select_from(j)
                .order_by(
                    workers_tbl.c.id,
                )
            )

            if _workerid is not None:
                q = q.where(workers_tbl.c.id == _workerid)
            if _name is not None:
                q = q.where(workers_tbl.c.name == _name)
            if masterid is not None:
                q = q.where(bm_tbl.c.masterid == masterid)
            if builderid is not None:
                q = q.where(bm_tbl.c.builderid == builderid)
            if paused is not None:
                q = q.where(workers_tbl.c.paused == int(paused))
            if graceful is not None:
                q = q.where(workers_tbl.c.graceful == int(graceful))

            rv: dict[int, WorkerModel] = {}
            res = None
            lastId = None
            for row in conn.execute(q):
                if row.id != lastId:
                    lastId = row.id
                    res = self._model_from_row(row)
                    rv[lastId] = res
                if row.builderid and row.masterid:
                    rv[lastId].configured_on.append(
                        BuilderMasterModel(builderid=row.builderid, masterid=row.masterid)
                    )

            # now go back and get the connection info for the same set of
            # workers
            j = conn_tbl
            if _name is not None:
                # note this is not an outer join; if there are unconnected
                # workers, they were captured in rv above
                j = j.join(workers_tbl)
            q = (
                sa.select(
                    conn_tbl.c.workerid,
                    conn_tbl.c.masterid,
                )
                .select_from(j)
                .order_by(conn_tbl.c.workerid)
                .where(conn_tbl.c.workerid.in_(rv.keys()))
            )

            if _name is not None:
                q = q.where(workers_tbl.c.name == _name)
            if masterid is not None:
                q = q.where(conn_tbl.c.masterid == masterid)

            for row in conn.execute(q):
                if row.workerid not in rv:
                    continue
                rv[row.workerid].connected_to.append(row.masterid)

            return list(rv.values())

        return self.db.pool.do(thd)

    # returns a Deferred that returns None
    def workerConnected(self, workerid, masterid, workerinfo):
        def thd(conn):
            conn_tbl = self.db.model.connected_workers
            q = conn_tbl.insert()
            try:
                conn.execute(q, {'workerid': workerid, 'masterid': masterid})
                conn.commit()
            except (sa.exc.IntegrityError, sa.exc.ProgrammingError):
                # if the row is already present, silently fail..
                conn.rollback()

            bs_tbl = self.db.model.workers
            q = bs_tbl.update().where(bs_tbl.c.id == workerid)
            conn.execute(q.values(info=workerinfo))
            conn.commit()

        return self.db.pool.do(thd)

    # returns a Deferred that returns None
    def workerDisconnected(self, workerid, masterid):
        def thd(conn):
            tbl = self.db.model.connected_workers
            q = tbl.delete().where(tbl.c.workerid == workerid, tbl.c.masterid == masterid)
            conn.execute(q)

        return self.db.pool.do_with_transaction(thd)

    # returns a Deferred that returns None
    def set_worker_paused(self, workerid, paused, pause_reason=None):
        def thd(conn):
            tbl = self.db.model.workers
            q = tbl.update().where(tbl.c.id == workerid)
            conn.execute(q.values(paused=int(paused), pause_reason=pause_reason))

        return self.db.pool.do_with_transaction(thd)

    # returns a Deferred that returns None
    def set_worker_graceful(self, workerid, graceful):
        def thd(conn):
            tbl = self.db.model.workers
            q = tbl.update().where(tbl.c.id == workerid)
            conn.execute(q.values(graceful=int(graceful)))

        return self.db.pool.do_with_transaction(thd)

    def _model_from_row(self, row):
        return WorkerModel(
            id=row.id,
            name=row.name,
            workerinfo=row.info,
            paused=bool(row.paused),
            pause_reason=row.pause_reason,
            graceful=bool(row.graceful),
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from typing import ClassVar
from typing import NamedTuple
from typing import Sequence

from zope.interface import implementer

from buildbot.interfaces import IRenderable
from buildbot.util import ComparableMixin
from buildbot.util.twisted import async_to_deferred


@implementer(IRenderable)
class GitCredentialInputRenderer(ComparableMixin):
    compare_attrs: ClassVar[Sequence[str]] = ('_credential_attributes',)

    def __init__(self, **credential_attributes) -> None:
        self._credential_attributes: dict[str, IRenderable | str] = credential_attributes

    @async_to_deferred
    async def getRenderingFor(self, build):
        props = build.getProperties()

        rendered_attributes = []

        attributes = list(self._credential_attributes.items())

        # git-credential-approve parsing of the `url` attribute
        # will reset all other fields
        # So make sure it's the first attribute in the form
        if 'url' in self._credential_attributes:
            attributes.sort(key=lambda e: e[0] != "url")

        for key, value in attributes:
            rendered_value = await props.render(value)
            if rendered_value is not None:
                rendered_attributes.append(f"{key}={rendered_value}\n")

        return "".join(rendered_attributes)


class GitCredentialOptions(NamedTuple):
    # Each element of `credentials` should be a `str` which is a input format for git-credential
    # ref: https://git-scm.com/docs/git-credential#IOFMT
    credentials: list[IRenderable | str]
    # value to set the git config `credential.useHttpPath` to.
    # ref: https://git-scm.com/docs/gitcredentials#Documentation/gitcredentials.txt-useHttpPath
    use_http_path: bool | None = None


def add_user_password_to_credentials(
    auth_credentials: tuple[IRenderable | str, IRenderable | str],
    url: IRenderable | str | None,
    credential_options: GitCredentialOptions | None,
) -> GitCredentialOptions:
    if credential_options is None:
        credential_options = GitCredentialOptions(credentials=[])
    else:
        # create a new instance to avoid side-effects
        credential_options = GitCredentialOptions(
            credentials=credential_options.credentials[:],
            use_http_path=credential_options.use_http_path,
        )

    username, password = auth_credentials
    credential_options.credentials.insert(
        0,
        IRenderable(  # placate typing
            GitCredentialInputRenderer(
                url=url,
                username=username,
                password=password,
            )
        ),
    )

    return credential_options



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

import os
import time
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import ClassVar
from typing import Sequence

from twisted.internet import defer
from twisted.python import log

from buildbot import config
from buildbot.changes import base
from buildbot.util import bytes2unicode
from buildbot.util import deferredLocked
from buildbot.util import runprocess
from buildbot.util.state import StateMixin

if TYPE_CHECKING:
    from buildbot.util.twisted import InlineCallbacksType


class HgPoller(base.ReconfigurablePollingChangeSource, StateMixin):
    """This source will poll a remote hg repo for changes and submit
    them to the change master."""

    compare_attrs: ClassVar[Sequence[str]] = (
        "repourl",
        "branch",
        "branches",
        "bookmarks",
        "workdir",
        "pollInterval",
        "hgpoller",
        "usetimestamps",
        "category",
        "project",
        "pollAtLaunch",
        "pollRandomDelayMin",
        "pollRandomDelayMax",
    )

    db_class_name = 'HgPoller'

    def __init__(self, repourl: str, **kwargs: Any) -> None:
        name = kwargs.get("name", None)
        if not name:
            branches = self.build_branches(kwargs.get('branch', None), kwargs.get('branches', None))
            kwargs["name"] = self.build_name(None, repourl, kwargs.get('bookmarks', None), branches)

        self.initLock = defer.DeferredLock()

        super().__init__(repourl, **kwargs)

    def checkConfig(  # type: ignore[override]
        self,
        repourl: str,
        branch: str | None = None,
        branches: list[str] | None = None,
        bookmarks: list[str] | None = None,
        workdir: str | None = None,
        pollInterval: int = 10 * 60,
        hgbin: str = "hg",
        usetimestamps: bool = True,
        category: str | Callable | None = None,
        project: str = "",
        encoding: str = "utf-8",
        name: str | None = None,
        pollAtLaunch: bool = False,
        revlink: Callable[[str, str], str] = lambda branch, revision: (""),
        pollRandomDelayMin: int = 0,
        pollRandomDelayMax: int = 0,
    ) -> None:
        # checkConfig doesn't need to match supertype
        if branch and branches:
            config.error("HgPoller: can't specify both branch and branches")

        if not callable(revlink):
            config.error("You need to provide a valid callable for revlink")

        if workdir is None:
            config.error("workdir is mandatory for now in HgPoller")

        name = self.build_name(name, repourl, bookmarks, self.build_branches(branch, branches))

        super().checkConfig(
            name=name,
            pollInterval=pollInterval,
            pollAtLaunch=pollAtLaunch,
            pollRandomDelayMin=pollRandomDelayMin,
            pollRandomDelayMax=pollRandomDelayMax,
        )

    @defer.inlineCallbacks
    def reconfigService(  # type: ignore[override]
        self,
        repourl: str,
        branch: str | None = None,
        branches: list[str] | None = None,
        bookmarks: list[str] | None = None,
        workdir: str | None = None,
        pollInterval: int = 10 * 60,
        hgbin: str = "hg",
        usetimestamps: bool = True,
        category: str | Callable | None = None,
        project: str = "",
        encoding: str = "utf-8",
        name: str | None = None,
        pollAtLaunch: bool = False,
        revlink: Callable[[str, str], str] = lambda branch, revision: (""),
        pollRandomDelayMin: int = 0,
        pollRandomDelayMax: int = 0,
    ) -> InlineCallbacksType[None]:
        # reconfigService doesn't need to match supertype
        self.repourl = repourl

        self.branches = self.build_branches(branch, branches)
        self.bookmarks = bookmarks or []

        name = self.build_name(name, repourl, bookmarks, self.branches)

        if not self.branches and not self.bookmarks:
            self.branches = ['default']

        self.encoding = encoding
        self.lastChange = time.time()
        self.lastPoll = time.time()
        self.hgbin = hgbin
        self.workdir = workdir
        self.usetimestamps = usetimestamps
        self.category = category if callable(category) else bytes2unicode(category)
        self.project = project
        self.lastRev: dict[str, str] = {}
        self.revlink_callable = revlink

        yield super().reconfigService(
            name=name,
            pollInterval=pollInterval,
            pollAtLaunch=pollAtLaunch,
            pollRandomDelayMin=pollRandomDelayMin,
            pollRandomDelayMax=pollRandomDelayMax,
        )

    def build_name(
        self,
        name: str | None,
        repourl: str,
        bookmarks: list[str] | None,
        branches: list[str] | None,
    ) -> str:
        if name is not None:
            return name

        name = repourl
        if bookmarks:
            name += "_" + "_".join(bookmarks)
        if branches:
            name += "_" + "_".join(branches)
        return name

    def build_branches(self, branch: str | None, branches: list[str] | None) -> list[str]:
        if branch:
            return [branch]
        return branches or []

    @defer.inlineCallbacks
    def activate(self) -> InlineCallbacksType[None]:
        self.lastRev = yield self.getState('lastRev', {})
        super().activate()

    def describe(self) -> str:
        status = ""
        if not self.master:
            status = "[STOPPED - check log]"
        return (
            f"HgPoller watching the remote Mercurial repository '{self.repourl}', "
            f"branches: {', '.join(self.branches)}, in workdir '{self.workdir}' {status}"
        )

    @deferredLocked('initLock')
    @defer.inlineCallbacks
    def poll(self) -> InlineCallbacksType[None]:  # type: ignore[override]
        yield self._getChanges()
        yield self._processChanges()

    def _absWorkdir(self) -> str:
        workdir = self.workdir
        assert workdir is not None
        if os.path.isabs(workdir):
            return workdir
        return os.path.join(self.master.basedir, workdir)

    @defer.inlineCallbacks
    def _getRevDetails(
        self, rev: str
    ) -> InlineCallbacksType[tuple[float | None, str, list[str], str]]:
        """Return a deferred for (date, author, files, comments) of given rev.

        Deferred will be in error if rev is unknown.
        """
        command = [
            self.hgbin,
            'log',
            '-r',
            rev,
            os.linesep.join((
                '--template={date|hgdate}',
                '{author}',
                "{files % '{file}" + os.pathsep + "'}",
                '{desc|strip}',
            )),
        ]

        # Mercurial fails with status 255 if rev is unknown
        rc, output = yield runprocess.run_process(
            self.master.reactor,
            command,
            workdir=self._absWorkdir(),
            env=os.environ,
            collect_stderr=False,
            stderr_is_error=True,
        )
        if rc != 0:
            msg = f'{self}: got error {rc} when getting details for revision {rev}'
            raise RuntimeError(msg)

        # all file names are on one line
        output = output.decode(self.encoding, "replace")
        date, author, files, comments = output.split(os.linesep, 3)

        if not self.usetimestamps:
            stamp = None
        else:
            try:
                stamp = float(date.split()[0])
            except Exception:
                log.msg(f'hgpoller: caught exception converting output {date!r} to timestamp')
                raise
        return stamp, author.strip(), files.split(os.pathsep)[:-1], comments.strip()

    def _isRepositoryReady(self) -> bool:
        """Easy to patch in tests."""
        return os.path.exists(os.path.join(self._absWorkdir(), '.hg'))

    @defer.inlineCallbacks
    def _initRepository(self) -> InlineCallbacksType[None]:
        """Have mercurial init the workdir as a repository (hg init) if needed.

        hg init will also create all needed intermediate directories.
        """
        if self._isRepositoryReady():
            return
        log.msg(f'hgpoller: initializing working dir from {self.repourl}')

        rc = yield runprocess.run_process(
            self.master.reactor,
            [self.hgbin, 'init', self._absWorkdir()],
            env=os.environ,
            collect_stdout=False,
            collect_stderr=False,
        )

        if rc != 0:
            self._stopOnFailure()
            raise OSError(f'{self}: repository init failed with exit code {rc}')

        log.msg(f"hgpoller: finished initializing working dir {self.workdir}")

    @defer.inlineCallbacks
    def _getChanges(self) -> InlineCallbacksType[None]:
        self.lastPoll = time.time()

        yield self._initRepository()
        log.msg(f"{self}: polling hg repo at {self.repourl}")

        command = [self.hgbin, 'pull']
        for name in self.branches:
            command += ['-b', name]
        for name in self.bookmarks:
            command += ['-B', name]
        command += [self.repourl]

        yield runprocess.run_process(
            self.master.reactor,
            command,
            workdir=self._absWorkdir(),
            env=os.environ,
            collect_stdout=False,
            collect_stderr=False,
        )

    def _getCurrentRev(self, branch: str = 'default') -> str | None:
        """Return a deferred for current numeric rev in state db.

        If never has been set, current rev is None.
        """
        return self.lastRev.get(branch, None)

    def _setCurrentRev(self, rev: str, branch: str = 'default') -> defer.Deferred[Any]:
        """Return a deferred to set current revision in persistent state."""
        self.lastRev[branch] = str(rev)
        return self.setState('lastRev', self.lastRev)

    @defer.inlineCallbacks
    def _getHead(self, branch: str) -> InlineCallbacksType[str | None]:
        """Return a deferred for branch head revision or None.

        We'll get an error if there is no head for this branch, which is
        probably a good thing, since it's probably a misspelling
        (if really buildbotting a branch that does not have any changeset
        yet, one shouldn't be surprised to get errors)
        """

        rc, stdout = yield runprocess.run_process(
            self.master.reactor,
            [self.hgbin, 'heads', branch, '--template={rev}' + os.linesep],
            workdir=self._absWorkdir(),
            env=os.environ,
            collect_stderr=False,
            stderr_is_error=True,
        )

        if rc != 0:
            log.err(f"{self}: could not find revision {branch} in repository {self.repourl}")
            return None

        if not stdout:
            return None

        if len(stdout.split()) > 1:
            log.err(
                f"{self}: caught several heads in branch {branch} "
                f"from repository {self.repourl}. Staying at previous revision"
                "You should wait until the situation is normal again "
                "due to a merge or directly strip if remote repo "
                "gets stripped later."
            )
            return None

        # in case of whole reconstruction, are we sure that we'll get the
        # same node -> rev assignations ?
        return stdout.strip().decode(self.encoding)

    @defer.inlineCallbacks
    def _processChanges(self) -> InlineCallbacksType[None]:
        """Send info about pulled changes to the master and record current.

        HgPoller does the recording by moving the working dir to the head
        of the branch.
        We don't update the tree (unnecessary treatment and waste of space)
        instead, we simply store the current rev number in a file.
        Recall that hg rev numbers are local and incremental.
        """
        for branch in self.branches + self.bookmarks:
            rev = yield self._getHead(branch)
            if rev is None:
                # Nothing pulled?
                continue
            yield self._processBranchChanges(rev, branch)

    @defer.inlineCallbacks
    def _getRevNodeList(self, revset: str) -> InlineCallbacksType[list[list[str]]]:
        rc, stdout = yield runprocess.run_process(
            self.master.reactor,
            [self.hgbin, 'log', '-r', revset, r'--template={rev}:{node}\n'],
            workdir=self._absWorkdir(),
            env=os.environ,
            collect_stdout=True,
            collect_stderr=False,
            stderr_is_error=True,
        )

        if rc != 0:
            raise OSError(f'{self}: could not get rev node list: {rc}')

        results = stdout.decode(self.encoding)

        revNodeList = [rn.split(':', 1) for rn in results.strip().split()]
        return revNodeList

    @defer.inlineCallbacks
    def _processBranchChanges(self, new_rev: str, branch: str) -> InlineCallbacksType[None]:
        prev_rev = yield self._getCurrentRev(branch)
        if new_rev == prev_rev:
            # Nothing new.
            return
        if prev_rev is None:
            # First time monitoring; start at the top.
            yield self._setCurrentRev(new_rev, branch)
            return

        # two passes for hg log makes parsing simpler (comments is multi-lines)
        revNodeList = yield self._getRevNodeList(f'{prev_rev}::{new_rev}')

        # revsets are inclusive. Strip the already-known "current" changeset.
        if not revNodeList:
            # empty revNodeList probably means the branch has changed head (strip of force push?)
            # in that case, we should still produce a change for that new rev (but we can't know
            # how many parents were pushed)
            revNodeList = yield self._getRevNodeList(new_rev)
        else:
            del revNodeList[0]

        log.msg(
            f'hgpoller: processing {len(revNodeList)} changes in branch '
            f'{branch!r}: {revNodeList!r} in {self._absWorkdir()!r}'
        )
        for _, node in revNodeList:
            timestamp, author, files, comments = yield self._getRevDetails(node)
            yield self.master.data.updates.addChange(
                author=author,
                committer=None,
                revision=str(node),
                revlink=self.revlink_callable(branch, str(node)),
                files=files,
                comments=comments,
                when_timestamp=int(timestamp) if timestamp else None,
                branch=bytes2unicode(branch),
                category=self.category,
                project=bytes2unicode(self.project),
                repository=bytes2unicode(self.repourl),
                src='hg',
            )
            # writing after addChange so that a rev is never missed,
            # but at once to avoid impact from later errors
            yield self._setCurrentRev(new_rev, branch)

    def _stopOnFailure(self) -> None:
        "utility method to stop the service when a failure occurs"
        if self.running:
            d = defer.maybeDeferred(self.stopService)
            d.addErrback(log.err, 'while stopping broken HgPoller service')



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import hashlib
import itertools
from typing import TYPE_CHECKING

import sqlalchemy as sa
from twisted.internet import defer

from buildbot.util import service
from buildbot.util import unicode2bytes
from buildbot.util.sautils import hash_columns

if TYPE_CHECKING:
    from buildbot.db.connector import DBConnector


class DBConnectorComponent(service.AsyncService):
    # A fixed component of the DBConnector, handling one particular aspect of
    # the database.  Instances of subclasses are assigned to attributes of the
    # DBConnector object, so that they are available at e.g.,
    # C{master.db.model} or C{master.db.changes}.  This parent class takes care
    # of the necessary backlinks and other housekeeping.

    connector: DBConnector | None = None
    data2db: dict[str, str] = {}

    def __init__(self, connector: DBConnector):
        self.db = connector

        # set up caches
        for method in dir(self.__class__):
            o = getattr(self, method)
            if isinstance(o, CachedMethod):
                setattr(self, method, o.get_cached_method(self))

    _isCheckLengthNecessary: bool | None = None

    def checkLength(self, col, value):
        if not self._isCheckLengthNecessary:
            if self.db.pool.engine.dialect.name == 'mysql':
                self._isCheckLengthNecessary = True
            else:
                # not necessary, so just stub out the method
                self.checkLength = lambda col, value: None
                return

        assert col.type.length, f"column {col} does not have a length"
        if value and len(value) > col.type.length:
            raise RuntimeError(
                f"value for column {col} is greater than max of {col.type.length} "
                f"characters: {value}"
            )

    def ensureLength(self, col, value):
        assert col.type.length, f"column {col} does not have a length"
        if value and len(value) > col.type.length:
            value = (
                value[: col.type.length // 2]
                + hashlib.sha1(unicode2bytes(value)).hexdigest()[: col.type.length // 2]
            )
        return value

    # returns a Deferred that returns a value
    @defer.inlineCallbacks
    def findSomethingId(self, tbl, whereclause, insert_values, _race_hook=None, autoCreate=True):
        pair = yield self.findOrCreateSomethingId(
            tbl, whereclause, insert_values, _race_hook, autoCreate
        )
        return pair[0]

    def findOrCreateSomethingId(
        self, tbl, whereclause, insert_values, _race_hook=None, autoCreate=True
    ):
        """
        Find a matching row and if one cannot be found optionally create it.
        Returns a deferred which resolves to the pair (id, found) where
        id is the primary key of the matching row and `found` is True if
        a match was found. `found` will be false if a new row was created.
        """

        def thd(conn, no_recurse=False):
            # try to find the master
            q = sa.select(tbl.c.id)
            if whereclause is not None:
                q = q.where(whereclause)
            r = conn.execute(q)
            row = r.fetchone()
            r.close()

            # found it!
            if row:
                return row.id, True

            if not autoCreate:
                return None, False

            if _race_hook is not None:
                _race_hook(conn)

            try:
                r = conn.execute(tbl.insert(), [insert_values])
                conn.commit()
                return r.inserted_primary_key[0], False
            except (sa.exc.IntegrityError, sa.exc.ProgrammingError):
                conn.rollback()
                # try it all over again, in case there was an overlapping,
                # identical call, but only retry once.
                if no_recurse:
                    raise
                return thd(conn, no_recurse=True)

        return self.db.pool.do(thd)

    def hashColumns(self, *args):
        return hash_columns(*args)

    def doBatch(self, batch, batch_n=500):
        iterator = iter(batch)
        while True:
            batch = list(itertools.islice(iterator, batch_n))
            if not batch:
                break
            yield batch


class CachedMethod:
    def __init__(self, cache_name, method):
        self.cache_name = cache_name
        self.method = method

    def get_cached_method(self, component):
        meth = self.method

        meth_name = meth.__name__
        cache = component.db.master.caches.get_cache(
            self.cache_name, lambda key: meth(component, key)
        )

        def wrap(key, no_cache=0):
            if no_cache:
                return meth(component, key)
            return cache.get(key)

        wrap.__name__ = meth_name + " (wrapped)"
        wrap.__module__ = meth.__module__
        wrap.__doc__ = meth.__doc__
        wrap.cache = cache
        return wrap


def cached(cache_name):
    return lambda method: CachedMethod(cache_name, method)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from buildbot.test.fakedb.row import Row


class Buildset(Row):
    table = "buildsets"

    id_column = 'id'

    def __init__(
        self,
        id=None,
        external_idstring='extid',
        reason='because',
        submitted_at=12345678,
        complete=0,
        complete_at=None,
        results=-1,
        rebuilt_buildid=None,
        parent_buildid=None,
        parent_relationship=None,
    ):
        super().__init__(
            id=id,
            external_idstring=external_idstring,
            reason=reason,
            submitted_at=submitted_at,
            complete=complete,
            complete_at=complete_at,
            results=results,
            rebuilt_buildid=rebuilt_buildid,
            parent_buildid=parent_buildid,
            parent_relationship=parent_relationship,
        )


class BuildsetProperty(Row):
    table = "buildset_properties"

    def __init__(self, buildsetid=None, property_name='prop', property_value='[22, "fakedb"]'):
        super().__init__(
            buildsetid=buildsetid, property_name=property_name, property_value=property_value
        )


class BuildsetSourceStamp(Row):
    table = "buildset_sourcestamps"

    id_column = 'id'

    def __init__(self, id=None, buildsetid=None, sourcestampid=None):
        super().__init__(id=id, buildsetid=buildsetid, sourcestampid=sourcestampid)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from collections import defaultdict
from collections import deque
from itertools import filterfalse
from weakref import WeakValueDictionary

from twisted.internet import defer
from twisted.python import log


class LRUCache:
    """
    A least-recently-used cache, with a fixed maximum size.

    See buildbot manual for more information.
    """

    __slots__ = (
        'max_size max_queue miss_fn queue cache weakrefs refcount hits refhits misses'.split()
    )
    sentinel = object()
    QUEUE_SIZE_FACTOR = 10

    def __init__(self, miss_fn, max_size=50):
        self.max_size = max_size
        self.max_queue = max_size * self.QUEUE_SIZE_FACTOR
        self.queue = deque()
        self.cache = {}
        self.weakrefs = WeakValueDictionary()
        self.hits = self.misses = self.refhits = 0
        self.refcount = defaultdict(lambda: 0)
        self.miss_fn = miss_fn

    def put(self, key, value):
        cached = key in self.cache or key in self.weakrefs
        self.cache[key] = value
        self.weakrefs[key] = value
        self._ref_key(key)
        if not cached:
            self._purge()

    def get(self, key, **miss_fn_kwargs):
        try:
            return self._get_hit(key)
        except KeyError:
            pass

        self.misses += 1

        result = self.miss_fn(key, **miss_fn_kwargs)
        if result is not None:
            self.cache[key] = result
            self.weakrefs[key] = result
            self._ref_key(key)
            self._purge()

        return result

    def keys(self):
        return list(self.cache)

    def set_max_size(self, max_size):
        if self.max_size == max_size:
            return

        self.max_size = max_size
        self.max_queue = max_size * self.QUEUE_SIZE_FACTOR
        self._purge()

    def inv(self):
        global inv_failed

        # the keys of the queue and cache should be identical
        cache_keys = set(self.cache.keys())
        queue_keys = set(self.queue)
        if queue_keys - cache_keys:
            log.msg("INV: uncached keys in queue:", queue_keys - cache_keys)
            inv_failed = True
        if cache_keys - queue_keys:
            log.msg("INV: unqueued keys in cache:", cache_keys - queue_keys)
            inv_failed = True

        # refcount should always represent the number of times each key appears
        # in the queue
        exp_refcount = {}
        for k in self.queue:
            exp_refcount[k] = exp_refcount.get(k, 0) + 1
        if exp_refcount != self.refcount:
            log.msg("INV: refcounts differ:")
            log.msg(" expected:", sorted(exp_refcount.items()))
            log.msg("      got:", sorted(self.refcount.items()))
            inv_failed = True

    def _ref_key(self, key):
        """Record a reference to the argument key."""
        queue = self.queue
        refcount = self.refcount

        queue.append(key)
        refcount[key] = refcount[key] + 1

        # periodically compact the queue by eliminating duplicate keys
        # while preserving order of most recent access.  Note that this
        # is only required when the cache does not exceed its maximum
        # size
        if len(queue) > self.max_queue:
            refcount.clear()
            queue_appendleft = queue.appendleft
            queue_appendleft(self.sentinel)
            for k in filterfalse(refcount.__contains__, iter(queue.pop, self.sentinel)):
                queue_appendleft(k)
                refcount[k] = 1

    def _get_hit(self, key):
        """Try to do a value lookup from the existing cache entries."""
        try:
            result = self.cache[key]
            self.hits += 1
            self._ref_key(key)
            return result
        except KeyError:
            pass

        result = self.weakrefs[key]
        self.refhits += 1
        self.cache[key] = result
        self._ref_key(key)
        return result

    def _purge(self):
        """
        Trim the cache down to max_size by evicting the
        least-recently-used entries.
        """
        if len(self.cache) <= self.max_size:
            return

        cache = self.cache
        refcount = self.refcount
        queue = self.queue
        max_size = self.max_size

        # purge least recently used entries, using refcount to count entries
        # that appear multiple times in the queue
        while len(cache) > max_size:
            refc = 1
            while refc:
                k = queue.popleft()
                refc = refcount[k] = refcount[k] - 1
            del cache[k]
            del refcount[k]


class AsyncLRUCache(LRUCache):
    """
    An LRU cache with asynchronous locking to ensure that in the common case of
    multiple concurrent requests for the same key, only one fetch is performed.
    """

    __slots__ = ['concurrent']

    def __init__(self, miss_fn, max_size=50):
        super().__init__(miss_fn, max_size=max_size)
        self.concurrent = {}

    def get(self, key, **miss_fn_kwargs):
        try:
            result = self._get_hit(key)
            return defer.succeed(result)
        except KeyError:
            pass

        concurrent = self.concurrent
        conc = concurrent.get(key)
        if conc:
            self.hits += 1
            d = defer.Deferred()
            conc.append(d)
            return d

        # if we're here, we've missed and need to fetch
        self.misses += 1

        # create a list of waiting deferreds for this key
        d = defer.Deferred()
        assert key not in concurrent
        concurrent[key] = [d]

        miss_d = self.miss_fn(key, **miss_fn_kwargs)

        def handle_result(result):
            if result is not None:
                self.cache[key] = result
                self.weakrefs[key] = result

                # reference the key once, possibly standing in for multiple
                # concurrent accesses
                self._ref_key(key)

                self._purge()

            # and fire all of the waiting Deferreds
            dlist = concurrent.pop(key)
            for d in dlist:
                d.callback(result)

        def handle_failure(f):
            # errback all of the waiting Deferreds
            dlist = concurrent.pop(key)
            for d in dlist:
                d.errback(f)

        miss_d.addCallbacks(handle_result, handle_failure)
        miss_d.addErrback(log.err)

        return d


# for tests
inv_failed = False



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import copy

from twisted.internet import defer


class CompatibleLatentWorkerMixin:
    builds_may_be_incompatible = True
    _actual_build_props = None

    def renderWorkerProps(self, build):
        # Deriving classes should implement this method to render and return
        # a Deferred that will have all properties that are needed to start a
        # worker as its result. The Deferred should result in data that can
        # be copied via copy.deepcopy
        #
        # During actual startup, renderWorkerPropsOnStart should be called
        # which will invoke renderWorkerProps, store a copy of the results for
        # later comparison and return them.
        raise NotImplementedError()

    @defer.inlineCallbacks
    def renderWorkerPropsOnStart(self, build):
        props = yield self.renderWorkerProps(build)
        self._actual_build_props = copy.deepcopy(props)
        return props

    def resetWorkerPropsOnStop(self):
        self._actual_build_props = None

    @defer.inlineCallbacks
    def isCompatibleWithBuild(self, build):
        if self._actual_build_props is None:
            return True

        requested_props = yield self.renderWorkerProps(build)

        return requested_props == self._actual_build_props



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
#
# We can't put this method in utility modules, because they import dependency packages


import datetime
import os
import re
from subprocess import PIPE
from subprocess import STDOUT
from subprocess import Popen


def gitDescribeToPep440(version):
    # git describe produce version in the form: v0.9.8-20-gf0f45ca
    # where 20 is the number of commit since last release, and gf0f45ca is the short commit id
    # preceded by 'g' we parse this a transform into a pep440 release version 0.9.9.dev20
    # (increment last digit and add dev before 20)

    VERSION_MATCH = re.compile(
        r'(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)(\.post(?P<post>\d+))?(-(?P<dev>\d+))?(-g(?P<commit>.+))?'
    )
    v = VERSION_MATCH.search(version)
    if v:
        major = int(v.group('major'))
        minor = int(v.group('minor'))
        patch = int(v.group('patch'))
        if v.group('dev'):
            patch += 1
            dev = int(v.group('dev'))
            return f"{major}.{minor}.{patch}.dev{dev}"
        if v.group('post'):
            return f"{major}.{minor}.{patch}.post{v.group('post')}"
        return f"{major}.{minor}.{patch}"

    return v


def mTimeVersion(init_file):
    cwd = os.path.dirname(os.path.abspath(init_file))
    m = 0
    for root, _, files in os.walk(cwd):
        for f in files:
            m = max(os.path.getmtime(os.path.join(root, f)), m)
    d = datetime.datetime.fromtimestamp(m, datetime.timezone.utc)
    return d.strftime("%Y.%m.%d")


def getVersionFromArchiveId(git_archive_id='$Format:%ct %(describe:abbrev=10)$'):
    """Extract the tag if a source is from git archive.

    When source is exported via `git archive`, the git_archive_id init value is modified
    and placeholders are expanded to the "archived" revision:

        %ct: committer date, UNIX timestamp
        %(describe:abbrev=10): git-describe output, always abbreviating to 10 characters of commit ID.
                               e.g. v3.10.0-850-g5bf957f89

    See man gitattributes(5) and git-log(1) (PRETTY FORMATS) for more details.
    """
    # mangle the magic string to make sure it is not replaced by git archive
    if not git_archive_id.startswith('$For' + 'mat:'):
        # source was modified by git archive, try to parse the version from
        # the value of git_archive_id

        tstamp, _, describe_output = git_archive_id.strip().partition(' ')
        if describe_output:
            # archived revision is tagged, use the tag
            return gitDescribeToPep440(describe_output)

        # archived revision is not tagged, use the commit date
        d = datetime.datetime.fromtimestamp(int(tstamp), datetime.timezone.utc)
        return d.strftime('%Y.%m.%d')
    return None


def getVersion(init_file):
    """
    Return BUILDBOT_VERSION environment variable, content of VERSION file, git
    tag or '0.0.0'  meaning we could not find the version, but the output still has to be valid
    """

    try:
        return os.environ['BUILDBOT_VERSION']
    except KeyError:
        pass

    try:
        cwd = os.path.dirname(os.path.abspath(init_file))
        fn = os.path.join(cwd, 'VERSION')
        with open(fn, encoding='utf-8') as f:
            return f.read().strip()
    except OSError:
        pass

    version = getVersionFromArchiveId()
    if version is not None:
        return version

    try:
        p = Popen(['git', 'describe', '--tags', '--always'], stdout=PIPE, stderr=STDOUT, cwd=cwd)
        out = p.communicate()[0]

        if (not p.returncode) and out:
            v = gitDescribeToPep440(str(out))
            if v:
                return v
    except OSError:
        pass

    try:
        # if we really can't find the version, we use the date of modification of the most recent
        # file docker hub builds cannot use git describe
        return mTimeVersion(init_file)
    except Exception:
        # bummer. lets report something
        return "0.0.0"


version = getVersion(__file__)

__version__ = version



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Portions Copyright Buildbot Team Members


from twisted.internet import protocol


class LineBuffer:
    def __init__(self):
        self._buffer = b''

    def add_data(self, data):
        # returns lines that have been processed, if any
        lines = (self._buffer + data).split(b'\n')
        self._buffer = lines.pop(-1)
        for l in lines:
            yield l.rstrip(b'\r')

    def get_trailing_line(self):
        if self._buffer:
            ret = [self._buffer]
            self._buffer = b''
            return ret
        return []


class LineProcessProtocol(protocol.ProcessProtocol):
    def __init__(self):
        self._out_buffer = LineBuffer()
        self._err_buffer = LineBuffer()

    def outReceived(self, data):
        """
        Translates bytes into lines, and calls outLineReceived.
        """
        for line in self._out_buffer.add_data(data):
            self.outLineReceived(line)

    def errReceived(self, data):
        """
        Translates bytes into lines, and calls errLineReceived.
        """
        for line in self._err_buffer.add_data(data):
            self.errLineReceived(line)

    def processEnded(self, reason):
        for line in self._out_buffer.get_trailing_line():
            self.outLineReceived(line)
        for line in self._err_buffer.get_trailing_line():
            self.errLineReceived(line)

    def outLineReceived(self, line):
        """
        Callback to which stdout lines will be sent.
        Any line that is not terminated by a newline will be processed once the next line comes,
        or when processEnded is called.
        """
        raise NotImplementedError

    def errLineReceived(self, line):
        """
        Callback to which stdout lines will be sent.
        Any line that is not terminated by a newline will be processed once the next line comes,
        or when processEnded is called.
        """
        raise NotImplementedError



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.trial import unittest


class VersioningUtilsTests(unittest.SynchronousTestCase):
    # Version utils are copied in three packages.
    # this unit test is made to be able to test the three versions
    # with the same test
    module_under_test = "buildbot"

    def setUp(self):
        try:
            self.m = __import__(self.module_under_test)
        except ImportError as e:
            raise unittest.SkipTest(self.module_under_test + " package is not installed") from e

    def test_gitDescribeToPep440devVersion(self):
        self.assertEqual(self.m.gitDescribeToPep440("v0.9.8-20-gf0f45ca"), "0.9.9.dev20")

    def test_gitDescribeToPep440tag(self):
        self.assertEqual(self.m.gitDescribeToPep440("v0.9.8"), "0.9.8")

    def test_gitDescribeToPep440p1tag(self):
        self.assertEqual(self.m.gitDescribeToPep440("v0.9.9.post1"), "0.9.9.post1")

    def test_gitDescribeToPep440p1dev(self):
        self.assertEqual(self.m.gitDescribeToPep440("v0.9.9.post1-20-gf0f45ca"), "0.9.10.dev20")

    def test_getVersionFromArchiveIdNoTag(self):
        version = self.m.getVersionFromArchiveId("1514651968 v0.9.9.post1-20-gf0f45ca")
        self.assertEqual(version, "0.9.10.dev20")

    def test_getVersionFromArchiveIdtag(self):
        version = self.m.getVersionFromArchiveId('1514808197 v1.0.0')
        self.assertEqual(version, "1.0.0")


class VersioningUtilsTests_PKG(VersioningUtilsTests):
    module_under_test = "buildbot_pkg"


class VersioningUtilsTests_WORKER(VersioningUtilsTests):
    module_under_test = "buildbot_worker"



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
"""
Miscellaneous utilities; these should be imported from C{buildbot.util}, not
directly from this module.
"""

from __future__ import annotations

import os
from functools import wraps
from typing import TYPE_CHECKING

from twisted.internet import reactor
from twisted.internet.defer import Deferred
from twisted.internet.defer import DeferredLock

if TYPE_CHECKING:
    from typing import Callable
    from typing import TypeVar

    from typing_extensions import ParamSpec

    _T = TypeVar('_T')
    _P = ParamSpec('_P')


def deferredLocked(
    lock_or_attr: str | DeferredLock,
) -> Callable[[Callable[_P, Deferred[_T]]], Callable[_P, Deferred[_T]]]:
    def decorator(fn: Callable[_P, Deferred[_T]]) -> Callable[_P, Deferred[_T]]:
        @wraps(fn)
        def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> Deferred[_T]:
            if isinstance(lock_or_attr, str):
                lock = getattr(args[0], lock_or_attr)
                assert isinstance(lock, DeferredLock)
            else:
                lock = lock_or_attr
            return lock.run(fn, *args, **kwargs)

        return wrapper

    return decorator


def cancelAfter(seconds, deferred, _reactor=reactor):
    delayedCall = _reactor.callLater(seconds, deferred.cancel)

    # cancel the delayedCall when the underlying deferred fires
    @deferred.addBoth
    def cancelTimer(x):
        if delayedCall.active():
            delayedCall.cancel()
        return x

    return deferred


def writeLocalFile(path, contents, mode=None):  # pragma: no cover
    with open(path, 'w', encoding='utf-8') as file:
        if mode is not None:
            os.chmod(path, mode)
        file.write(contents)


def chunkify_list(l, chunk_size):
    chunk_size = max(1, chunk_size)
    return (l[i : i + chunk_size] for i in range(0, len(l), chunk_size))



#!/usr/bin/env python
#
# Copyright 2008
# Steve 'Ashcrow' Milner <smilner+buildbot@redhat.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


import os
import sys

"""
Generates changelog information using git.
"""

__docformat__ = 'restructuredtext'


def print_err(msg):
    """
    Wrapper to make printing to stderr nicer.

    :Parameters:
       - `msg`: the message to print.
    """
    sys.stderr.write(msg)
    sys.stderr.write('\n')


def usage():
    """
    Prints out usage information to stderr.
    """
    print_err(f'Usage: {sys.argv[0]} git-binary since')
    print_err(f'Example: {sys.argv[0]} /usr/bin/git f5067523dfae9c7cdefc828721ec593ac7be62db')


def main(args):
    """
    Main entry point.

    :Parameters:
       - `args`: same as sys.argv[1:]
    """
    # Make sure we have the arguments we need, else show usage
    try:
        git_bin = args[0]
        since = args[1]
    except IndexError:
        usage()
        return 1

    if not os.access(git_bin, os.X_OK):
        print_err(f'Can not access {git_bin}')
        return 1

    # Open a pipe and force the format
    pipe = os.popen(git_bin + ' log --pretty="format:%ad  %ae%n  * %s" ' + since + '..')
    print(pipe.read())
    pipe.close()
    return 0


if __name__ == '__main__':
    raise SystemExit(main(sys.argv[1:]))



#!/usr/bin/env python
"""Change source forwarder for bitbucket.org POST service.

bitbucket_buildbot.py will determine the repository information from
the JSON HTTP POST it receives from bitbucket.org and build the
appropriate repository.

If your bitbucket repository is private, you must add a ssh key to the
bitbucket repository for the user who initiated bitbucket_buildbot.py

bitbucket_buildbot.py is based on github_buildbot.py
"""

import logging
import sys
import tempfile
import traceback
from optparse import OptionParser

from future.utils import iteritems
from twisted.cred import credentials
from twisted.internet import reactor
from twisted.spread import pb
from twisted.web import resource
from twisted.web import server

try:
    import json
except ImportError:
    import simplejson as json


class BitBucketBuildBot(resource.Resource):
    """
    BitBucketBuildBot creates the webserver that responds to the
    BitBucket POST Service Hook.
    """

    isLeaf = True
    bitbucket = None
    master = None
    port = None
    private = False

    def render_POST(self, request):
        """
        Responds only to POST events and starts the build process

        :arguments:
            request
                the http request object
        """
        try:
            payload = json.loads(request.args['payload'][0])
            logging.debug("Payload: " + str(payload))
            self.process_change(payload)
        except Exception:
            logging.error("Encountered an exception:")
            for msg in traceback.format_exception(*sys.exc_info()):
                logging.error(msg.strip())

    def process_change(self, payload):
        """
        Consumes the JSON as a python object and actually starts the build.

        :arguments:
            payload
                Python Object that represents the JSON sent by Bitbucket POST
                Service Hook.
        """
        if self.private:
            repo_url = 'ssh://hg@{}{}'.format(
                self.bitbucket,
                payload['repository']['absolute_url'],
            )
        else:
            repo_url = 'http://{}{}'.format(
                self.bitbucket,
                payload['repository']['absolute_url'],
            )
        changes = []
        for commit in payload['commits']:
            files = [file_info['file'] for file_info in commit['files']]
            revlink = 'http://{}{}/changeset/{}/'.format(
                self.bitbucket,
                payload['repository']['absolute_url'],
                commit['node'],
            )
            change = {
                'revision': commit['node'],
                'revlink': revlink,
                'comments': commit['message'],
                'who': commit['author'],
                'files': files,
                'repository': repo_url,
                'properties': dict(),
            }
            changes.append(change)
        # Submit the changes, if any
        if not changes:
            logging.warning("No changes found")
            return
        host, port = self.master.split(':')
        port = int(port)
        factory = pb.PBClientFactory()
        deferred = factory.login(credentials.UsernamePassword("change", "changepw"))
        logging.debug('Trying to connect to: %s:%d', host, port)
        reactor.connectTCP(host, port, factory)
        deferred.addErrback(self.connectFailed)
        deferred.addCallback(self.connected, changes)

    def connectFailed(self, error):
        """
        If connection is failed.  Logs the error.
        """
        logging.error("Could not connect to master: %s", error.getErrorMessage())
        return error

    def addChange(self, dummy, remote, changei, src='hg'):
        """
        Sends changes from the commit to the buildmaster.
        """
        logging.debug("addChange %s, %s", repr(remote), repr(changei))
        try:
            change = changei.next()
        except StopIteration:
            remote.broker.transport.loseConnection()
            return None

        logging.info("New revision: %s", change['revision'][:8])
        for key, value in iteritems(change):
            logging.debug("  %s: %s", key, value)

        change['src'] = src
        deferred = remote.callRemote('addChange', change)
        deferred.addCallback(self.addChange, remote, changei, src)
        return deferred

    def connected(self, remote, changes):
        """
        Responds to the connected event.
        """
        return self.addChange(None, remote, changes.__iter__())


def main():
    """
    The main event loop that starts the server and configures it.
    """
    usage = "usage: %prog [options]"
    parser = OptionParser(usage)
    parser.add_option(
        "-p",
        "--port",
        help="Port the HTTP server listens to for the Bitbucket Service Hook [default: %default]",
        default=4000,
        type=int,
        dest="port",
    )
    parser.add_option(
        "-m",
        "--buildmaster",
        help="Buildbot Master host and port. ie: localhost:9989 [default:" + " %default]",
        default="localhost:9989",
        dest="buildmaster",
    )
    parser.add_option(
        "-l",
        "--log",
        help="The absolute path, including filename, to save the log to [default: %default]",
        default=tempfile.gettempdir() + "/bitbucket_buildbot.log",
        dest="log",
    )
    parser.add_option(
        "-L",
        "--level",
        help="The logging level: debug, info, warn, error, fatal [default: %default]",
        default='warn',
        dest="level",
    )
    parser.add_option(
        "-g",
        "--bitbucket",
        help="The bitbucket serve [default: %default]",
        default='bitbucket.org',
        dest="bitbucket",
    )
    parser.add_option(
        '-P',
        '--private',
        help='Use SSH to connect, for private repositories.',
        dest='private',
        default=False,
        action='store_true',
    )
    (options, _) = parser.parse_args()
    # Set up logging.
    levels = {
        'debug': logging.DEBUG,
        'info': logging.INFO,
        'warn': logging.WARNING,
        'error': logging.ERROR,
        'fatal': logging.FATAL,
    }
    filename = options.log
    log_format = "%(asctime)s - %(levelname)s - %(message)s"
    logging.basicConfig(filename=filename, format=log_format, level=levels[options.level])
    # Start listener.
    bitbucket_bot = BitBucketBuildBot()
    bitbucket_bot.bitbucket = options.bitbucket
    bitbucket_bot.master = options.buildmaster
    bitbucket_bot.private = options.private
    site = server.Site(bitbucket_bot)
    reactor.listenTCP(options.port, site)
    reactor.run()


if __name__ == '__main__':
    main()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer
from twisted.internet import reactor

from buildbot import config
from buildbot.process.buildstep import FAILURE
from buildbot.process.buildstep import SUCCESS
from buildbot.process.buildstep import BuildStep

# use the 'requests' lib: https://requests.readthedocs.io/en/master/
try:
    import requests
    import txrequests
except ImportError:
    txrequests = None


# This step uses a global Session object, which encapsulates a thread pool as
# well as state such as cookies and authentication.  This state may pose
# problems for users, where one step may get a cookie that is subsequently used
# by another step in a different build.

_session = None


def getSession():
    global _session
    if _session is None:
        _session = txrequests.Session()
        reactor.addSystemEventTrigger("before", "shutdown", closeSession)
    return _session


def setSession(session):
    global _session
    _session = session


def closeSession():
    global _session
    if _session is not None:
        _session.close()
        _session = None


def _headerSet(headers):
    return frozenset(map(lambda x: x.casefold(), headers))


class HTTPStep(BuildStep):
    name = 'HTTPStep'
    description = 'Requesting'
    descriptionDone = 'Requested'
    requestsParams = [
        "params",
        "data",
        "json",
        "headers",
        "cookies",
        "files",
        "auth",
        "timeout",
        "allow_redirects",
        "proxies",
        "hooks",
        "stream",
        "verify",
        "cert",
    ]
    renderables = [*requestsParams, "method", "url"]
    session = None

    def __init__(
        self, url, method, hide_request_headers=None, hide_response_headers=None, **kwargs
    ):
        if txrequests is None:
            config.error("Need to install txrequest to use this step:\n\n pip install txrequests")

        if method not in ('POST', 'GET', 'PUT', 'DELETE', 'HEAD', 'OPTIONS'):
            config.error(f"Wrong method given: '{method}' is not known")

        self.method = method
        self.url = url

        self.hide_request_headers = _headerSet(hide_request_headers or [])
        self.hide_response_headers = _headerSet(hide_response_headers or [])

        for param in self.requestsParams:
            setattr(self, param, kwargs.pop(param, None))

        super().__init__(**kwargs)

    @defer.inlineCallbacks
    def run(self):
        # create a new session if it doesn't exist
        self.session = getSession()

        requestkwargs = {'method': self.method, 'url': self.url}

        for param in self.requestsParams:
            value = getattr(self, param, None)
            if value is not None:
                requestkwargs[param] = value

        log = yield self.addLog('log')

        # known methods already tested in __init__

        yield log.addHeader(f'Performing {self.method} request to {self.url}\n')
        if self.params:
            yield log.addHeader('Parameters:\n')
            params = sorted(self.params.items(), key=lambda x: x[0])
            requestkwargs['params'] = params
            for k, v in params:
                yield log.addHeader(f'\t{k}: {v}\n')
        data = requestkwargs.get("data", None)
        if data:
            yield log.addHeader('Data:\n')
            if isinstance(data, dict):
                for k, v in data.items():
                    yield log.addHeader(f'\t{k}: {v}\n')
            else:
                yield log.addHeader(f'\t{data}\n')

        try:
            r = yield self.session.request(**requestkwargs)
        except requests.exceptions.ConnectionError as e:
            yield log.addStderr(f'An exception occurred while performing the request: {e}')
            return FAILURE

        if r.history:
            yield log.addStdout(f'\nRedirected {len(r.history)} times:\n\n')
            for rr in r.history:
                yield self.log_response(log, rr)
                yield log.addStdout('=' * 60 + '\n')

        yield self.log_response(log, r)

        yield log.finish()

        self.descriptionDone = [f"Status code: {r.status_code}"]
        if r.status_code < 400:
            return SUCCESS
        else:
            return FAILURE

    @defer.inlineCallbacks
    def log_response(self, log, response):
        yield log.addHeader('Request Headers:\n')
        for k, v in response.request.headers.items():
            if k.casefold() in self.hide_request_headers:
                v = '<HIDDEN>'
            yield log.addHeader(f'\t{k}: {v}\n')

        yield log.addStdout(f'URL: {response.url}\n')

        if response.status_code == requests.codes.ok:
            yield log.addStdout(f'Status: {response.status_code}\n')
        else:
            yield log.addStderr(f'Status: {response.status_code}\n')

        yield log.addHeader('Response Headers:\n')
        for k, v in response.headers.items():
            if k.casefold() in self.hide_response_headers:
                v = '<HIDDEN>'
            yield log.addHeader(f'\t{k}: {v}\n')

        yield log.addStdout(f' ------ Content ------\n{response.text}')
        content_log = yield self.addLog('content')
        yield content_log.addStdout(response.text)


class POST(HTTPStep):
    def __init__(self, url, **kwargs):
        super().__init__(url, method='POST', **kwargs)


class GET(HTTPStep):
    def __init__(self, url, **kwargs):
        super().__init__(url, method='GET', **kwargs)


class PUT(HTTPStep):
    def __init__(self, url, **kwargs):
        super().__init__(url, method='PUT', **kwargs)


class DELETE(HTTPStep):
    def __init__(self, url, **kwargs):
        super().__init__(url, method='DELETE', **kwargs)


class HEAD(HTTPStep):
    def __init__(self, url, **kwargs):
        super().__init__(url, method='HEAD', **kwargs)


class OPTIONS(HTTPStep):
    def __init__(self, url, **kwargs):
        super().__init__(url, method='OPTIONS', **kwargs)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import defer

from buildbot.test.util.integration import RunMasterBase


# This integration test creates a master and worker environment,
# with one builder and a shellcommand step
# meant to be a template for integration steps
class ShellMaster(RunMasterBase):
    @defer.inlineCallbacks
    def setup_config(self):
        c = {}
        from buildbot.config import BuilderConfig
        from buildbot.plugins import schedulers
        from buildbot.plugins import steps
        from buildbot.process.factory import BuildFactory

        c['schedulers'] = [
            schedulers.AnyBranchScheduler(name="sched", builderNames=["testy"]),
            schedulers.ForceScheduler(name="force", builderNames=["testy"]),
        ]

        f = BuildFactory()
        f.addStep(steps.ShellCommand(command='echo hello'))
        c['builders'] = [BuilderConfig(name="testy", workernames=["local1"], factory=f)]
        yield self.setup_master(c)

    @defer.inlineCallbacks
    def test_shell(self):
        yield self.setup_config()
        # if you don't need change, you can just remove this change, and useChange parameter
        change = {
            "branch": "master",
            "files": ["foo.c"],
            "author": "me@foo.com",
            "committer": "me@foo.com",
            "comments": "good stuff",
            "revision": "HEAD",
            "project": "none",
        }
        build = yield self.doForceBuild(
            wantSteps=True, useChange=change, wantLogs=True, wantProperties=True
        )
        self.assertEqual(build['buildid'], 1)
        self.assertEqual(build['properties']['owners'], (['me@foo.com'], 'Build'))



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from io import StringIO

from twisted.internet import defer

from buildbot.test.util.integration import RunMasterBase
from buildbot.test.util.integration import print_build

# This integration test creates a master and worker environment,
# with two builders and a trigger step linking them

expectedOutputRegex = r"""\*\*\* BUILD 1 \*\*\* ==> build successful \(success\)
    \*\*\* STEP worker_preparation \*\*\* ==> worker local1 ready \(success\)
    \*\*\* STEP shell \*\*\* ==> 'echo hello' \(success\)
        log:stdio \({loglines}\)
    \*\*\* STEP trigger \*\*\* ==> triggered trigsched, 1 success \(success\)
       url:trigsched #2 \(http://localhost:8080/#/buildrequests/2\)
       url:success: build #1 \(http://localhost:8080/#/builders/(1|2)/builds/1\)
    \*\*\* STEP shell_1 \*\*\* ==> 'echo world' \(success\)
        log:stdio \({loglines}\)
\*\*\* BUILD 2 \*\*\* ==> build successful \(success\)
    \*\*\* STEP worker_preparation \*\*\* ==> worker local1 ready \(success\)
    \*\*\* STEP shell \*\*\* ==> 'echo ola' \(success\)
        log:stdio \({loglines}\)
"""


class TriggeringMaster(RunMasterBase):
    change = {
        "branch": "master",
        "files": ["foo.c"],
        "author": "me@foo.com",
        "committer": "me@foo.com",
        "comments": "good stuff",
        "revision": "HEAD",
        "project": "none",
    }

    @defer.inlineCallbacks
    def setup_config(self, addFailure=False):
        c = {}
        from buildbot.config import BuilderConfig
        from buildbot.plugins import schedulers
        from buildbot.plugins import steps
        from buildbot.process.factory import BuildFactory

        c['schedulers'] = [
            schedulers.Triggerable(name="trigsched", builderNames=["build"]),
            schedulers.AnyBranchScheduler(name="sched", builderNames=["testy"]),
        ]

        f = BuildFactory()
        f.addStep(steps.ShellCommand(command='echo hello'))
        f.addStep(
            steps.Trigger(schedulerNames=['trigsched'], waitForFinish=True, updateSourceStamp=True)
        )
        f.addStep(steps.ShellCommand(command='echo world'))
        f2 = BuildFactory()
        f2.addStep(steps.ShellCommand(command='echo ola'))
        c['builders'] = [
            BuilderConfig(name="testy", workernames=["local1"], factory=f),
            BuilderConfig(name="build", workernames=["local1"], factory=f2),
        ]
        if addFailure:
            f3 = BuildFactory()
            f3.addStep(steps.ShellCommand(command='false'))
            c['builders'].append(BuilderConfig(name="build2", workernames=["local1"], factory=f3))
            c['builders'].append(BuilderConfig(name="build3", workernames=["local1"], factory=f2))
            c['schedulers'][0] = schedulers.Triggerable(
                name="trigsched", builderNames=["build", "build2", "build3"]
            )
        yield self.setup_master(c)

    @defer.inlineCallbacks
    def test_trigger(self):
        yield self.setup_config()

        build = yield self.doForceBuild(wantSteps=True, useChange=self.change, wantLogs=True)

        self.assertEqual(build['steps'][2]['state_string'], 'triggered trigsched, 1 success')
        builds = yield self.master.data.get(("builds",))
        self.assertEqual(len(builds), 2)
        dump = StringIO()
        for b in builds:
            yield print_build(b, self.master, dump)
        # depending on the environment the number of lines is different between
        # test hosts
        loglines = builds[1]['steps'][1]['logs'][0]['num_lines']
        self.assertRegex(dump.getvalue(), expectedOutputRegex.format(loglines=loglines))

    @defer.inlineCallbacks
    def test_trigger_failure(self):
        yield self.setup_config(addFailure=True)

        build = yield self.doForceBuild(wantSteps=True, useChange=self.change, wantLogs=True)

        self.assertEqual(
            build['steps'][2]['state_string'], 'triggered trigsched, 2 successes, 1 failure'
        )
        builds = yield self.master.data.get(("builds",))
        self.assertEqual(len(builds), 4)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from __future__ import annotations

from buildbot.data import base
from buildbot.data import types

# NOTE: patches are not available via endpoints


class Patch(base.ResourceType):
    name = "patch"
    plural = "patches"
    endpoints: list[type[base.Endpoint]] = []

    class EntityType(types.Entity):
        patchid = types.Integer()
        body = types.Binary()
        level = types.Integer()
        subdir = types.String()
        author = types.String()
        comment = types.String()

    entityType = EntityType(name)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from buildbot.process import buildstep
from buildbot.process.results import SUCCESS
from buildbot.statistics import capture
from buildbot.statistics.storage_backends.base import StatsStorageBase


class FakeStatsStorageService(StatsStorageBase):
    """
    Fake Storage service used in unit tests
    """

    def __init__(self, stats=None, name='FakeStatsStorageService'):
        self.stored_data = []
        if not stats:
            self.stats = [capture.CaptureProperty("TestBuilder", 'test')]
        else:
            self.stats = stats
        self.name = name
        self.captures = []

    def thd_postStatsValue(self, post_data, series_name, context=None):
        if not context:
            context = {}
        self.stored_data.append((post_data, series_name, context))


class FakeBuildStep(buildstep.BuildStep):
    """
    A fake build step to be used for testing.
    """

    def doSomething(self):
        self.setProperty("test", 10, "test")

    def start(self):
        self.doSomething()
        return SUCCESS


class FakeInfluxDBClient:
    """
    Fake Influx module for testing on systems that don't have influxdb installed.
    """

    def __init__(self, *args, **kwargs):
        self.points = []

    def write_points(self, points):
        self.points.extend(points)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from typing import ClassVar
from typing import Sequence

from buildbot import util
from buildbot.config.checks import check_param_str


class Codebase(util.ComparableMixin):
    compare_attrs: ClassVar[Sequence[str]] = (
        "name",
        "slug",
        "project",
    )

    def __init__(self, name: str, project: str, slug: str | None = None) -> None:
        if slug is None:
            slug = name

        self.name = check_param_str(name, self.__class__, "name")
        self.slug = check_param_str(slug, self.__class__, "slug")
        self.project = check_param_str(project, self.__class__, "project")



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
"""
file based provider
"""

from __future__ import annotations

import os
import stat

from buildbot import config
from buildbot.secrets.providers.base import SecretProviderBase


class SecretInAFile(SecretProviderBase):
    """
    secret is stored in a separate file under the given directory name
    """

    name: str | None = "SecretInAFile"  # type: ignore[assignment]

    def checkFileIsReadOnly(self, dirname, secretfile):
        filepath = os.path.join(dirname, secretfile)
        obs_stat = stat.S_IMODE(os.stat(filepath).st_mode)
        if (obs_stat & 0o7) != 0 and os.name == "posix":
            config.error(
                f"Permissions {oct(obs_stat)} on file {secretfile} are too open."
                " It is required that your secret files are NOT"
                " accessible by others!"
            )

    def checkSecretDirectoryIsAvailableAndReadable(self, dirname, suffixes):
        if not os.access(dirname, os.F_OK):
            config.error(f"directory {dirname} does not exists")
        for secretfile in os.listdir(dirname):
            for suffix in suffixes:
                if secretfile.endswith(suffix):
                    self.checkFileIsReadOnly(dirname, secretfile)

    def loadSecrets(self, dirname, suffixes, strip):
        secrets = {}
        for secretfile in os.listdir(dirname):
            secretvalue = None
            for suffix in suffixes:
                if secretfile.endswith(suffix):
                    with open(os.path.join(dirname, secretfile), encoding='utf-8') as source:
                        secretvalue = source.read()
                    if suffix:
                        secretfile = secretfile[: -len(suffix)]
                    if strip:
                        secretvalue = secretvalue.rstrip("\r\n")
                    secrets[secretfile] = secretvalue
        return secrets

    def checkConfig(self, dirname, suffixes=None, strip=True):
        self._dirname = dirname
        if suffixes is None:
            suffixes = [""]
        self.checkSecretDirectoryIsAvailableAndReadable(dirname, suffixes=suffixes)

    def reconfigService(self, dirname, suffixes=None, strip=True):
        self._dirname = dirname
        self.secrets = {}
        if suffixes is None:
            suffixes = [""]
        self.secrets = self.loadSecrets(self._dirname, suffixes=suffixes, strip=strip)

    def get(self, entry):
        """
        get the value from the file identified by 'entry'
        """
        return self.secrets.get(entry)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from typing import TYPE_CHECKING

from twisted.internet import defer

from buildbot.data import base
from buildbot.data import patches
from buildbot.data import types

if TYPE_CHECKING:
    from typing import Any

    from buildbot.db.sourcestamps import SourceStampModel


def _db2data(ss: SourceStampModel):
    data: dict[str, Any] = {
        'ssid': ss.ssid,
        'branch': ss.branch,
        'revision': ss.revision,
        'project': ss.project,
        'repository': ss.repository,
        'codebase': ss.codebase,
        'created_at': ss.created_at,
        'patch': None,
    }
    if ss.patch is not None:
        data['patch'] = {
            'patchid': ss.patch.patchid,
            'level': ss.patch.level,
            'subdir': ss.patch.subdir,
            'author': ss.patch.author,
            'comment': ss.patch.comment,
            'body': ss.patch.body,
        }
    return data


class SourceStampEndpoint(base.Endpoint):
    kind = base.EndpointKind.SINGLE
    pathPatterns = [
        "/sourcestamps/n:ssid",
    ]

    @defer.inlineCallbacks
    def get(self, resultSpec, kwargs):
        ssdict = yield self.master.db.sourcestamps.getSourceStamp(kwargs['ssid'])
        return _db2data(ssdict) if ssdict else None


class SourceStampsEndpoint(base.Endpoint):
    kind = base.EndpointKind.COLLECTION
    pathPatterns = [
        "/sourcestamps",
        "/buildsets/:buildsetid/sourcestamps",
    ]
    rootLinkName = 'sourcestamps'

    @defer.inlineCallbacks
    def get(self, resultSpec, kwargs):
        buildsetid = kwargs.get("buildsetid")
        if buildsetid is not None:
            sourcestamps = yield self.master.db.sourcestamps.get_sourcestamps_for_buildset(
                buildsetid
            )
        else:
            sourcestamps = yield self.master.db.sourcestamps.getSourceStamps()

        return [_db2data(ssdict) for ssdict in sourcestamps]


class SourceStamp(base.ResourceType):
    name = "sourcestamp"
    plural = "sourcestamps"
    endpoints = [SourceStampEndpoint, SourceStampsEndpoint]

    class EntityType(types.Entity):
        ssid = types.Integer()
        revision = types.NoneOk(types.String())
        branch = types.NoneOk(types.String())
        repository = types.String()
        project = types.String()
        codebase = types.String()
        patch = types.NoneOk(patches.Patch.entityType)
        created_at = types.DateTime()

    entityType = EntityType(name)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import contextlib
import os
import re
from typing import TYPE_CHECKING
from typing import Any
from typing import ClassVar
from typing import Sequence
from typing import cast
from urllib.parse import quote as urlquote

from twisted.internet import defer
from twisted.python import log

from buildbot import config
from buildbot.changes import base
from buildbot.process.codebase import Codebase
from buildbot.util import bytes2unicode
from buildbot.util import giturlparse
from buildbot.util import private_tempdir
from buildbot.util import runprocess
from buildbot.util import unicode2bytes
from buildbot.util.git import GitMixin
from buildbot.util.git import GitServiceAuth
from buildbot.util.git import check_ssh_config
from buildbot.util.git_credential import GitCredentialOptions
from buildbot.util.git_credential import add_user_password_to_credentials
from buildbot.util.state import StateMixin
from buildbot.util.twisted import InlineCallbacksType
from buildbot.util.twisted import async_to_deferred

if TYPE_CHECKING:
    from typing import Callable
    from typing import Literal

    from buildbot.interfaces import IRenderable


class GitError(Exception):
    """Raised when git exits with code 128."""


class GitPoller(base.ReconfigurablePollingChangeSource, StateMixin, GitMixin):
    """This source will poll a remote git repo for changes and submit
    them to the change master."""

    compare_attrs: ClassVar[Sequence[str]] = (
        "repourl",
        "branches",
        "workdir",
        "pollInterval",
        "gitbin",
        "usetimestamps",
        "category",
        'codebase',
        "project",
        "pollAtLaunch",
        "buildPushesWithNoCommits",
        "pollRandomDelayMin",
        "pollRandomDelayMax",
        "_git_auth",
    )

    def __init__(self, repourl: str, **kwargs: Any) -> None:
        self._git_auth = GitServiceAuth(self)

        self.lastRev: dict[str, str] | None = None

        name = kwargs.get("name", None)
        if name is None:
            kwargs["name"] = repourl
        super().__init__(repourl, **kwargs)

    def checkConfig(  # type: ignore[override]
        self,
        repourl: str,
        branches: list[str] | Literal[True] | Callable[[str], bool] | None = None,
        branch: str | None = None,
        workdir: str | None = None,
        pollInterval: int = 10 * 60,
        gitbin: str = "git",
        usetimestamps: bool = True,
        category: str | Callable[[str], str] | None = None,
        codebase: Codebase | None = None,
        project: str | None = None,
        fetch_refspec: str | None = None,
        encoding: str = "utf-8",
        name: str | None = None,
        pollAtLaunch: bool = False,
        buildPushesWithNoCommits: bool = False,
        only_tags: bool = False,
        sshPrivateKey: str | None = None,
        sshHostKey: str | None = None,
        sshKnownHosts: str | None = None,
        pollRandomDelayMin: int = 0,
        pollRandomDelayMax: int = 0,
        auth_credentials: tuple[IRenderable | str, IRenderable | str] | None = None,
        git_credentials: GitCredentialOptions | None = None,
    ) -> None:
        if only_tags and (branch or branches):
            config.error("GitPoller: can't specify only_tags and branch/branches")
        if branch and branches:
            config.error("GitPoller: can't specify both branch and branches")
        if branch and not isinstance(branch, str):
            config.error("GitPoller: 'branch' argument must be a str")
        if branches is not None and not (
            (isinstance(branches, list) and all(isinstance(e, str) for e in branches))
            or branches is True
            or callable(branches)
        ):
            config.error(
                "GitPoller: 'branches' argument must be one of "
                "list of str, True, or Callable[[str], bool]"
            )

        check_ssh_config('GitPoller', sshPrivateKey, sshHostKey, sshKnownHosts)

        if fetch_refspec is not None:
            config.error(
                "GitPoller: fetch_refspec is no longer supported. "
                "Instead, only the given branches are downloaded."
            )

        if name is None:
            name = repourl

        if not isinstance(codebase, (Codebase, type(None))):
            config.error(
                f'{self.__class__.__name__}: codebase must be None or instance of Codebase'
            )

        super().checkConfig(
            name=name,
            pollInterval=pollInterval,
            pollAtLaunch=pollAtLaunch,
            pollRandomDelayMin=pollRandomDelayMin,
            pollRandomDelayMax=pollRandomDelayMax,
        )

    @defer.inlineCallbacks
    def reconfigService(  # type: ignore[override]
        self,
        repourl: str,
        branches: list[str] | Literal[True] | Callable[[str], bool] | None = None,
        branch: str | None = None,
        workdir: str | None = None,
        pollInterval: int = 10 * 60,
        gitbin: str = "git",
        usetimestamps: bool = True,
        category: str | Callable[[str], str] | None = None,
        codebase: Codebase | None = None,
        project: str | None = None,
        fetch_refspec: str | None = None,
        encoding: str = "utf-8",
        name: str | None = None,
        pollAtLaunch: bool = False,
        buildPushesWithNoCommits: bool = False,
        only_tags: bool = False,
        sshPrivateKey: str | None = None,
        sshHostKey: str | None = None,
        sshKnownHosts: str | None = None,
        pollRandomDelayMin: int = 0,
        pollRandomDelayMax: int = 0,
        auth_credentials: tuple[IRenderable | str, IRenderable | str] | None = None,
        git_credentials: GitCredentialOptions | None = None,
    ) -> InlineCallbacksType[None]:
        if name is None:
            name = repourl

        if project is None:
            project = ''

        if branch:
            branches = [branch]
        elif not branches:
            if only_tags:
                branches = lambda ref: ref.startswith('refs/tags/')
            else:
                branches = None

        self.repourl = repourl
        self.branches = branches
        self.encoding = encoding
        self.buildPushesWithNoCommits = buildPushesWithNoCommits
        self.gitbin = gitbin
        self.workdir = workdir
        self.usetimestamps = usetimestamps
        self.category = (
            category if callable(category) else bytes2unicode(category, encoding=self.encoding)
        )
        self.codebase = codebase

        if codebase is not None:
            projectid = yield self.master.data.updates.find_project_id(
                codebase.project, auto_create=False
            )
            if projectid is None:
                raise RuntimeError(f'Project {codebase.project} is not configured')

            self._codebase_id = yield self.master.data.updates.find_codebase_id(
                projectid=projectid, name=codebase.name
            )
        else:
            self._codebase_id = None

        self.project = bytes2unicode(project, encoding=self.encoding)
        self.lastRev = None

        self.setupGit()

        if auth_credentials is not None:
            git_credentials = add_user_password_to_credentials(
                auth_credentials,
                repourl,
                git_credentials,
            )

        self._git_auth = GitServiceAuth(
            self, sshPrivateKey, sshHostKey, sshKnownHosts, git_credentials
        )

        if self.workdir is None:
            self.workdir = 'gitpoller-work'

        # make our workdir absolute, relative to the master's basedir

        if not os.path.isabs(self.workdir):
            self.workdir = os.path.join(self.master.basedir, self.workdir)
            log.msg(f"gitpoller: using workdir '{self.workdir}'")

        yield super().reconfigService(
            name=name,
            pollInterval=pollInterval,
            pollAtLaunch=pollAtLaunch,
            pollRandomDelayMin=pollRandomDelayMin,
            pollRandomDelayMax=pollRandomDelayMax,
        )

    @defer.inlineCallbacks
    def _checkGitFeatures(self) -> InlineCallbacksType[None]:
        stdout = yield self._dovccmd('--version', [])

        self.parseGitFeatures(stdout)
        if not self.gitInstalled:
            raise OSError('Git is not installed')

        if not self.supportsSshPrivateKeyAsEnvOption:
            has_ssh_private_key = (
                yield self.renderSecrets(self._git_auth.ssh_private_key)
            ) is not None
            if has_ssh_private_key:
                raise OSError('SSH private keys require Git 2.3.0 or newer')

    def activate(self) -> defer.Deferred[None]:  # type: ignore[override]
        try:
            self.lastRev = None

            super().activate()
        except Exception as e:
            log.err(e, 'while initializing GitPoller repository')
        return defer.succeed(None)

    def describe(self) -> str:
        str = 'GitPoller watching the remote git repository ' + bytes2unicode(
            self.repourl, self.encoding
        )

        if self.branches:
            if self.branches is True:
                str += ', branches: ALL'
            elif not callable(self.branches):
                str += ', branches: ' + ', '.join(self.branches)

        if not self.master:
            str += " [STOPPED - check log]"

        return str

    @async_to_deferred
    async def _resolve_head_ref(self, git_auth_files_path: str | None = None) -> str | None:
        if self.supports_lsremote_symref:
            rows: str = await self._dovccmd(
                'ls-remote',
                ['--symref', self.repourl, 'HEAD'],
                auth_files_path=git_auth_files_path,
            )
            # simple parse of output which should have format:
            # ref: refs/heads/{branch}	HEAD
            # {hash}	HEAD
            parts = rows.split(maxsplit=3)
            # sanity just in case
            if len(parts) >= 3 and parts[0] == 'ref:' and parts[2] == 'HEAD':
                return parts[1]
            return None

        # naive fallback if git version does not support --symref
        rows = await self._dovccmd('ls-remote', [self.repourl, 'HEAD', 'refs/heads/*'])
        refs = [row.split('\t') for row in rows.splitlines() if '\t' in row]
        # retrieve hash that HEAD points to
        head_hash = next((hash for hash, ref in refs if ref == 'HEAD'), None)
        if head_hash is None:
            return None

        # get refs that points to the same hash as HEAD
        candidates = [ref for hash, ref in refs if ref != 'HEAD' and hash == head_hash]
        # Found default branch
        if len(candidates) == 1:
            return candidates[0]

        # If multiple ref points to the same hash as HEAD,
        # we have no way to know which one is the default
        return None

    @async_to_deferred
    async def _list_remote_refs(
        self, refs: list[str] | None = None, git_auth_files_path: str | None = None
    ) -> list[str]:
        rows: str = await self._dovccmd(
            'ls-remote',
            ['--refs', self.repourl] + (refs if refs is not None else []),
            auth_files_path=git_auth_files_path,
        )

        branches: list[str] = []
        for row in rows.splitlines():
            if '\t' not in row:
                # Not a useful line
                continue
            _, ref = row.split("\t")
            branches.append(ref)

        return branches

    @staticmethod
    def _trim_prefix(value: str, prefix: str) -> str:
        """Remove prefix from value."""
        if value.startswith(prefix):
            return value[len(prefix) :]
        return value

    def _removeHeads(self, branch: str) -> str:
        """Remove 'refs/heads/' prefix from remote references."""
        if branch.startswith("refs/heads/"):
            branch = branch[11:]
        return branch

    @staticmethod
    def _tracker_ref(repourl: str, ref: str) -> str:
        def _sanitize(value: str) -> str:
            return urlquote(value, '').replace('~', '%7E')

        tracker_prefix = "refs/buildbot"
        # if ref is not a Git ref, store under a different path to avoid collision
        if not ref.startswith('refs/'):
            tracker_prefix += "/raw"

        git_url = giturlparse(repourl)
        if git_url is None:
            # fallback to using the whole repourl
            url_identifier = _sanitize(repourl)
        else:
            url_identifier = f"{git_url.proto}/{_sanitize(git_url.domain)}"
            if git_url.port is not None:
                # replace `:` with url encode `%3A`
                url_identifier += f"%3A{git_url.port}"

            if git_url.owner is not None:
                url_identifier += f"/{_sanitize(git_url.owner)}"
            url_identifier += f"/{_sanitize(git_url.repo)}"

        return f"{tracker_prefix}/{url_identifier}/{GitPoller._trim_prefix(ref, 'refs/')}"

    def poll_should_exit(self) -> bool:
        # A single gitpoller loop may take a while on a loaded master, which would block
        # reconfiguration, so we try to exit early.
        return not self.doPoll.running

    @defer.inlineCallbacks
    def poll(self) -> InlineCallbacksType[None]:  # type: ignore[override]
        yield self._checkGitFeatures()

        try:
            assert self.workdir is not None
            yield self._dovccmd('init', ['--bare', self.workdir])
        except GitError as e:
            log.msg(e.args[0])
            return

        tmp_dir = (
            private_tempdir.PrivateTemporaryDirectory(dir=self.workdir, prefix='.buildbot-ssh')
            if self._git_auth.is_auth_needed
            else contextlib.nullcontext()
        )
        # retrieve auth files
        with tmp_dir as tmp_path:
            yield self._git_auth.download_auth_files_if_needed(cast(str, tmp_path))

            refs, trim_ref_head = yield self._get_refs(cast(str, tmp_path))

            # Nothing to fetch and process.
            if not refs:
                return

            if self.poll_should_exit():
                return

            refspecs = [f'+{ref}:{self._tracker_ref(self.repourl, ref)}' for ref in refs]

            try:
                yield self._dovccmd(
                    'fetch',
                    ["--progress", self.repourl, *refspecs, "--"],
                    path=self.workdir,
                    auth_files_path=tmp_path,
                )
            except GitError as e:
                log.msg(e.args[0])
                return

        if self.lastRev is None:
            self.lastRev = yield self.getState('lastRev', {})

        revs = {}
        log.msg(f'gitpoller: processing changes from "{self.repourl}"')
        for ref in refs:
            branch = ref if not trim_ref_head else self._trim_prefix(ref, 'refs/heads/')
            try:
                if self.poll_should_exit():  # pragma: no cover
                    # Note that we still want to update the last known revisions for the branches
                    # we did process
                    break

                rev = yield self._dovccmd(
                    'rev-parse', [self._tracker_ref(self.repourl, ref)], path=self.workdir
                )
                revs[branch] = rev
                yield self._process_changes(rev, branch)
            except Exception:
                log.err(_why=f"trying to poll branch {branch} of {self.repourl}")

        self.lastRev = revs
        yield self.setState('lastRev', self.lastRev)

    @async_to_deferred
    async def _get_refs(self, git_auth_files_path: str) -> tuple[list[str], bool]:
        if callable(self.branches):
            # Get all refs and let callback filter them
            remote_refs = await self._list_remote_refs(git_auth_files_path=git_auth_files_path)
            refs = [b for b in remote_refs if self.branches(b)]
            return (refs, False)

        if self.branches is True:
            # Get all branch refs
            refs = await self._list_remote_refs(
                refs=["refs/heads/*"],
                git_auth_files_path=git_auth_files_path,
            )
            return (refs, False)

        if self.branches:
            refs = await self._list_remote_refs(
                refs=[f"refs/heads/{b}" for b in self.branches],
                git_auth_files_path=git_auth_files_path,
            )
            return (refs, True)

        head_ref = await self._resolve_head_ref(git_auth_files_path=git_auth_files_path)
        if head_ref is not None:
            return ([head_ref], False)

        # unlikely, but if we can't find HEAD here, something weird happen,
        # but not a critical error. Just use HEAD as the ref to use
        return (['HEAD'], False)

    def _get_commit_comments(self, rev: str) -> defer.Deferred[str]:
        args = ['--no-walk', r'--format=%s%n%b', rev, '--']
        d = self._dovccmd('log', args, path=self.workdir)
        return d  # type: ignore[return-value]

    @defer.inlineCallbacks
    def _get_commit_timestamp(self, rev: str) -> InlineCallbacksType[int | None]:
        # unix timestamp
        args = ['--no-walk', r'--format=%ct', rev, '--']
        git_output = yield self._dovccmd('log', args, path=self.workdir)
        if self.usetimestamps:
            try:
                stamp = int(git_output)
            except Exception as e:
                log.msg(
                    f'gitpoller: caught exception converting output \'{git_output}\' to timestamp'
                )
                raise e
            return stamp
        return None

    @defer.inlineCallbacks
    def _get_commit_files(self, rev: str) -> InlineCallbacksType[list[str]]:
        args = ['--name-only', '--no-walk', r'--format=%n', '-m', '--first-parent', rev, '--']
        git_output = yield self._dovccmd('log', args, path=self.workdir)

        def decode_file(file: str) -> str:
            # git use octal char sequences in quotes when non ASCII
            match = re.match('^"(.*)"$', file)
            if match:
                file = bytes2unicode(
                    match.groups()[0], encoding=self.encoding, errors='unicode_escape'
                )
            return bytes2unicode(file, encoding=self.encoding)

        fileList = [decode_file(file) for file in [s for s in git_output.splitlines() if len(s)]]
        return fileList

    @defer.inlineCallbacks
    def _get_commit_author(self, rev: str) -> InlineCallbacksType[str]:
        args = ['--no-walk', r'--format=%aN <%aE>', rev, '--']
        git_output = yield self._dovccmd('log', args, path=self.workdir)
        if not git_output:
            raise OSError('could not get commit author for rev')
        return git_output

    @defer.inlineCallbacks
    def _get_commit_committer(self, rev: str) -> InlineCallbacksType[str]:
        args = ['--no-walk', r'--format=%cN <%cE>', rev, '--']
        res = yield self._dovccmd('log', args, path=self.workdir)
        if not res:
            raise OSError('could not get commit committer for rev')
        return res

    def _get_commit_parent_hashes(self, rev: str) -> defer.Deferred[str]:
        args = ['--no-walk', r'--format=%P', rev, '--']
        d = self._dovccmd('log', args, path=self.workdir)
        return d

    @defer.inlineCallbacks
    def _process_changes(self, newRev: str, branch: str) -> InlineCallbacksType[None]:
        """
        Read changes since last change.

        - Read list of commit hashes.
        - Extract details from each commit.
        - Add changes to database.
        """

        # initial run, don't parse all history
        if not self.lastRev:
            return

        # get the change list
        revListArgs = (
            ['--ignore-missing', '--first-parent']
            + ['--format=%H', f'{newRev}']
            + ['^' + rev for rev in sorted(self.lastRev.values())]
            + ['--']
        )
        results = yield self._dovccmd('log', revListArgs, path=self.workdir)

        # process oldest change first
        revList = results.split()
        revList.reverse()

        if self.buildPushesWithNoCommits and not revList:
            existingRev = self.lastRev.get(branch)
            if existingRev != newRev:
                revList = [newRev]
                if existingRev is None:
                    # This branch was completely unknown, rebuild
                    log.msg(f'gitpoller: rebuilding {newRev} for new branch "{branch}"')
                else:
                    # This branch is known, but it now points to a different
                    # commit than last time we saw it, rebuild.
                    log.msg(f'gitpoller: rebuilding {newRev} for updated branch "{branch}"')

        change_count = len(revList)
        self.lastRev[branch] = newRev

        if change_count:
            log.msg(
                f'gitpoller: processing {change_count} changes: {revList} from '
                f'"{self.repourl}" branch "{branch}"'
            )

        last_commit_id = None
        if self._codebase_id is not None and change_count:
            rev = revList[0]
            parent_hashes = yield self._get_commit_parent_hashes(rev)
            parent_hash = parent_hashes.split()[0]
            last_commit = yield self.master.data.get((
                'codebases',
                self._codebase_id,
                'commits_by_revision',
                parent_hash,
            ))
            if last_commit is not None:
                last_commit_id = last_commit['commitid']

        for rev in revList:
            dl: defer.Deferred[Any] = defer.DeferredList(
                [
                    self._get_commit_timestamp(rev),
                    self._get_commit_author(rev),
                    self._get_commit_committer(rev),
                    self._get_commit_files(rev),
                    self._get_commit_comments(rev),
                ],
                consumeErrors=True,
            )

            results = yield dl

            # check for failures
            failures = [r[1] for r in results if not r[0]]
            if failures:
                for failure in failures:
                    log.err(failure, f"while processing changes for {rev} {branch}")
                # just fail on the first error; they're probably all related!
                failures[0].raiseException()

            timestamp, author, committer, files, comments = [r[1] for r in results]

            yield self.master.data.updates.addChange(
                author=author,
                committer=committer,
                revision=bytes2unicode(rev, encoding=self.encoding),
                files=files,
                comments=comments,
                when_timestamp=timestamp,
                branch=bytes2unicode(self._removeHeads(branch)),
                project=self.project,
                repository=bytes2unicode(self.repourl, encoding=self.encoding),
                category=self.category,
                src='git',
            )

            if self._codebase_id is not None:
                last_commit_id = yield self.master.data.updates.add_commit(
                    codebaseid=self._codebase_id,
                    author=author,
                    committer=committer,
                    comments=comments,
                    when_timestamp=timestamp,
                    revision=bytes2unicode(rev, encoding=self.encoding),
                    parent_commitid=last_commit_id,
                )

        if self._codebase_id is not None and last_commit_id is not None:
            yield self.master.data.updates.update_branch(
                codebaseid=self._codebase_id,
                name=branch,
                commitid=last_commit_id,
                last_timestamp=int(self.master.reactor.seconds()),
            )

    @async_to_deferred
    async def _dovccmd(
        self,
        command: str,
        args: list[str],
        path: str | None = None,
        auth_files_path: str | None = None,
        initial_stdin: str | None = None,
    ) -> str:
        full_args: list[str] = []
        full_env = os.environ.copy()

        if self._git_auth.is_auth_needed_for_git_command(command):
            if auth_files_path is None:
                raise RuntimeError(
                    f"Git command {command} requires auth, but no auth information was provided"
                )
            self._git_auth.adjust_git_command_params_for_auth(
                full_args,
                full_env,
                auth_files_path,
                self,
            )

        full_args += [command, *args]

        res = await runprocess.run_process(
            self.master.reactor,
            [self.gitbin, *full_args],
            path,
            env=full_env,
            initial_stdin=unicode2bytes(initial_stdin) if initial_stdin is not None else None,
        )
        (code, stdout, stderr) = res
        stdout = bytes2unicode(stdout, self.encoding)
        stderr = bytes2unicode(stderr, self.encoding)
        if code != 0:
            if code == 128:
                raise GitError(
                    f'command {full_args} in {path} on repourl {self.repourl} failed '
                    f'with exit code {code}: {stderr}'
                )
            raise OSError(
                f'command {full_args} in {path} on repourl {self.repourl} '
                f'failed with exit code {code}: {stderr}'
            )
        return stdout.strip()



# This file is part of Buildbot. Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer
from twisted.logger import Logger

from buildbot.interfaces import LatentWorkerFailedToSubstantiate
from buildbot.util import asyncSleep
from buildbot.util import httpclientservice
from buildbot.util import kubeclientservice
from buildbot.util.latent import CompatibleLatentWorkerMixin
from buildbot.worker.docker import DockerBaseWorker

log = Logger()


class KubeError(RuntimeError):
    pass


class KubeJsonError(KubeError):
    def __init__(self, code, response_json):
        super().__init__(response_json['message'])
        self.code = code
        self.json = response_json
        self.reason = response_json.get('reason')


class KubeTextError(KubeError):
    def __init__(self, code, response):
        super().__init__(response)
        self.code = code


class KubeLatentWorker(CompatibleLatentWorkerMixin, DockerBaseWorker):
    instance = None
    _namespace = None
    _kube = None
    _kube_config = None

    @defer.inlineCallbacks
    def getPodSpec(self, build):
        image = yield build.render(self.image)
        env = yield self.createEnvironment(build)

        return {
            "apiVersion": "v1",
            "kind": "Pod",
            "metadata": {"name": self.getContainerName()},
            "spec": {
                "affinity": (yield self.get_affinity(build)),
                "containers": [
                    {
                        "name": self.getContainerName(),
                        "image": image,
                        "env": [{"name": k, "value": v} for k, v in env.items()],
                        "resources": (yield self.getBuildContainerResources(build)),
                        "volumeMounts": (yield self.get_build_container_volume_mounts(build)),
                    }
                ]
                + (yield self.getServicesContainers(build)),
                "nodeSelector": (yield self.get_node_selector(build)),
                "restartPolicy": "Never",
                "volumes": (yield self.get_volumes(build)),
            },
        }

    def getBuildContainerResources(self, build):
        # customization point to generate Build container resources
        return {}

    def get_build_container_volume_mounts(self, build):
        return []

    def get_affinity(self, build):
        return {}

    def get_node_selector(self, build):
        return {}

    def get_volumes(self, build):
        return []

    def getServicesContainers(self, build):
        # customization point to create services containers around the build container
        # those containers will run within the same localhost as the build container (aka within
        # the same pod)
        return []

    def renderWorkerProps(self, build_props):
        return self.getPodSpec(build_props)

    def checkConfig(
        self,
        name,
        image='buildbot/buildbot-worker',
        namespace=None,
        masterFQDN=None,
        master_protocol='pb',
        kube_config=None,
        **kwargs,
    ):
        super().checkConfig(name, None, master_protocol=master_protocol, **kwargs)

    @defer.inlineCallbacks
    def reconfigService(
        self,
        name,
        image='buildbot/buildbot-worker',
        namespace=None,
        masterFQDN=None,
        master_protocol='pb',
        kube_config=None,
        **kwargs,
    ):
        # Set build_wait_timeout to 0 if not explicitly set: Starting a
        # container is almost immediate, we can afford doing so for each build.
        if 'build_wait_timeout' not in kwargs:
            kwargs['build_wait_timeout'] = 0
        if masterFQDN is None:
            masterFQDN = self.get_ip
        if callable(masterFQDN):
            masterFQDN = masterFQDN()

        self._http = yield httpclientservice.HTTPSession(
            self.master.httpservice, kube_config.get_master_url()
        )

        if self.running and self._kube is not None:
            yield self._kube.unregister(self)

        self._kube = yield kubeclientservice.KubeClientService.getService(self.master)
        self._kube_config = kube_config

        if self.running:
            yield self._kube.register(self, kube_config)

        self._namespace = namespace or kube_config.getConfig()['namespace']

        yield super().reconfigService(
            name, image=image, masterFQDN=masterFQDN, master_protocol=master_protocol, **kwargs
        )

    @defer.inlineCallbacks
    def startService(self):
        yield super().startService()
        yield self._kube.register(self, self._kube_config)

    @defer.inlineCallbacks
    def stopService(self):
        yield self._kube.unregister(self)
        yield super().stopService()

    @defer.inlineCallbacks
    def start_instance(self, build):
        try:
            yield self.stop_instance(reportFailure=False)
            pod_spec = yield self.renderWorkerPropsOnStart(build)
            yield self._create_pod(self._namespace, pod_spec)
        except KubeError as e:
            raise LatentWorkerFailedToSubstantiate(str(e)) from e
        return True

    @defer.inlineCallbacks
    def stop_instance(self, fast=False, reportFailure=True):
        self.current_pod_spec = None
        self.resetWorkerPropsOnStop()
        try:
            yield self._delete_pod(self._namespace, self.getContainerName())
        except KubeJsonError as e:
            if reportFailure and e.reason != 'NotFound':
                raise
        if fast:
            return
        yield self._wait_for_pod_deletion(
            self._namespace, self.getContainerName(), timeout=self.missing_timeout
        )

    @defer.inlineCallbacks
    def _get_request_kwargs(self):
        config = self._kube_config.getConfig()

        kwargs = {}

        if config.get("headers"):
            kwargs.setdefault("headers", {}).update(config["headers"])

        auth = yield self._kube_config.getAuthorization()
        if auth is not None:
            kwargs.setdefault("headers", {})['Authorization'] = auth

        # warning: this only works with txrequests! not treq
        for arg in ['cert', 'verify']:
            if arg in config:
                kwargs[arg] = config[arg]

        return kwargs

    @defer.inlineCallbacks
    def _raise_decode_failure_error(self, res):
        content = yield res.content()
        msg = "Failed to decode: " + content.decode("utf-8", errors="ignore")[0:200]
        raise KubeTextError(res.code, msg)

    @defer.inlineCallbacks
    def _create_pod(self, namespace, spec):
        url = f'/api/v1/namespaces/{namespace}/pods'
        res = yield self._http.post(url, json=spec, **(yield self._get_request_kwargs()))

        try:
            res_json = yield res.json()
        except Exception:
            yield self._raise_decode_failure_error(res)

        if res.code not in (200, 201, 202):
            raise KubeJsonError(res.code, res_json)
        return res_json

    @defer.inlineCallbacks
    def _delete_pod(self, namespace, name, graceperiod=0):
        url = f'/api/v1/namespaces/{namespace}/pods/{name}'
        res = yield self._http.delete(
            url, params={'graceperiod': graceperiod}, **(yield self._get_request_kwargs())
        )

        try:
            res_json = yield res.json()
        except Exception:
            yield self._raise_decode_failure_error(res)

        if res.code != 200:
            raise KubeJsonError(res.code, res_json)
        return res_json

    @defer.inlineCallbacks
    def _wait_for_pod_deletion(self, namespace, name, timeout):
        t1 = self.master.reactor.seconds()
        url = f'/api/v1/namespaces/{namespace}/pods/{name}/status'
        while True:
            if self.master.reactor.seconds() - t1 > timeout:
                raise TimeoutError(f"Did not see pod {name} terminate after {timeout}s")
            res = yield self._http.get(url, **(yield self._get_request_kwargs()))

            try:
                res_json = yield res.json()
            except Exception:
                yield self._raise_decode_failure_error(res)

            if res.code == 404:
                break  # 404 means the pod has terminated
            if res.code != 200:
                raise KubeJsonError(res.code, res_json)
            yield asyncSleep(1, reactor=self.master.reactor)
        return res_json



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import sys

from twisted.web import static

from buildbot.util import bytes2unicode

if sys.version_info[:2] >= (3, 9):
    # We need importlib.resources.files, which is added in Python 3.9
    # https://docs.python.org/3/library/importlib.resources.html
    import importlib.resources as importlib_resources
else:
    import importlib_resources  # type: ignore[import-not-found]


class Application:
    def __init__(self, package_name: str, description: str, ui: bool = True) -> None:
        self.description = description
        self.version = importlib_resources.files(package_name).joinpath("VERSION")
        self.version = bytes2unicode(self.version.read_bytes())
        self.static_dir = importlib_resources.files(package_name) / "static"
        self.resource = static.File(self.static_dir)
        self.ui = ui

    def setMaster(self, master: object) -> None:
        self.master = master

    def setConfiguration(self, config: object) -> None:
        self.config = config

    def __repr__(self) -> str:
        return (
            "www.plugin.Application(version={version}, "
            "description={description}, "
            "static_dir={static_dir})"
        ).format(**self.__dict__)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import re

from twisted.python import log


class LoggingMixin:
    def setUpLogging(self):
        self._logEvents = []
        log.addObserver(self._logEvents.append)
        self.addCleanup(log.removeObserver, self._logEvents.append)

    def logContainsMessage(self, regexp):
        r = re.compile(regexp)
        for event in self._logEvents:
            msg = log.textFromEventDict(event)
            if msg is not None:
                assert not msg.startswith("Unable to format event"), msg
            if msg is not None and r.search(msg):
                return True
        return False

    def assertLogged(self, regexp):
        if not self.logContainsMessage(regexp):
            lines = [log.textFromEventDict(e) for e in self._logEvents]
            self.fail(f"{regexp!r} not matched in log output.\n{lines} ")

    def assertNotLogged(self, regexp):
        if self.logContainsMessage(regexp):
            lines = [log.textFromEventDict(e) for e in self._logEvents]
            self.fail(f"{regexp!r} matched in log output.\n{lines} ")

    def assertWasQuiet(self):
        self.assertEqual([log.textFromEventDict(event) for event in self._logEvents], [])



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
#
# This code has originally been copied from cpython project.
#
# Copyright Python Software Foundation and contributors
# Licensed under Python Software Foundation License Version 2

import ntpath
import os
import posixpath


def posix_expanduser(path, worker_environ):
    """Expand ~ and ~user constructions.  If user or $HOME is unknown,
    do nothing."""
    path = os.fspath(path)
    tilde = '~'
    if not path.startswith(tilde):
        return path
    sep = posixpath._get_sep(path)
    i = path.find(sep, 1)
    if i < 0:
        i = len(path)
    if i == 1:
        if 'HOME' not in worker_environ:
            try:
                import pwd
            except ImportError:
                # pwd module unavailable, return path unchanged
                return path
            try:
                userhome = pwd.getpwuid(os.getuid()).pw_dir
            except KeyError:
                # bpo-10496: if the current user identifier doesn't exist in the
                # password database, return the path unchanged
                return path
        else:
            userhome = worker_environ['HOME']
    else:
        try:
            import pwd
        except ImportError:
            # pwd module unavailable, return path unchanged
            return path
        name = path[1:i]
        try:
            pwent = pwd.getpwnam(name)
        except KeyError:
            # bpo-10496: if the user name from the path doesn't exist in the
            # password database, return the path unchanged
            return path
        userhome = pwent.pw_dir
    root = '/'
    userhome = userhome.rstrip(root)
    return (userhome + path[i:]) or root


def nt_expanduser(path, worker_environ):
    """Expand ~ and ~user constructs.
    If user or $HOME is unknown, do nothing."""
    path = os.fspath(path)
    tilde = '~'
    if not path.startswith(tilde):
        return path
    i = 1
    n = len(path)
    while i < n and path[i] not in ntpath._get_bothseps(path):
        i += 1

    if 'USERPROFILE' in worker_environ:
        userhome = worker_environ['USERPROFILE']
    elif 'HOMEPATH' not in worker_environ:
        return path
    else:
        try:
            drive = worker_environ['HOMEDRIVE']
        except KeyError:
            drive = ''
        userhome = ntpath.join(drive, worker_environ['HOMEPATH'])

    if i != 1:  # ~user
        target_user = path[1:i]
        current_user = worker_environ.get('USERNAME')

        if target_user != current_user:
            # Try to guess user home directory.  By default all user
            # profile directories are located in the same place and are
            # named by corresponding usernames.  If userhome isn't a
            # normal profile directory, this guess is likely wrong,
            # so we bail out.
            if current_user != ntpath.basename(userhome):
                return path
            userhome = ntpath.join(ntpath.dirname(userhome), target_user)

    return userhome + path[i:]



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING

import sqlalchemy as sa
from twisted.python import deprecate
from twisted.python import versions

from buildbot.db import base
from buildbot.util import identifiers
from buildbot.warnings import warn_deprecated

if TYPE_CHECKING:
    from twisted.internet import defer


@dataclasses.dataclass
class UserModel:
    uid: int
    identifier: str
    bb_username: str | None = None
    bb_password: str | None = None
    attributes: dict[str, str] | None = None

    # For backward compatibility
    def __getitem__(self, key: str):
        warn_deprecated(
            '4.1.0',
            (
                'UsersConnectorComponent '
                'getUser, getUserByUsername, and getUsers '
                'no longer return User as dictionnaries. '
                'Usage of [] accessor is deprecated: please access the member directly'
            ),
        )

        if hasattr(self, key):
            return getattr(self, key)

        if self.attributes is not None and key in self.attributes:
            return self.attributes[key]

        raise KeyError(key)


@deprecate.deprecated(versions.Version("buildbot", 4, 1, 0), UserModel)
class UsDict(UserModel):
    pass


class UsersConnectorComponent(base.DBConnectorComponent):
    def findUserByAttr(
        self, identifier: str, attr_type: str, attr_data: str, _race_hook=None
    ) -> defer.Deferred[int]:
        # note that since this involves two tables, self.findSomethingId is not
        # helpful
        def thd(conn, no_recurse=False, identifier=identifier) -> int:
            tbl = self.db.model.users
            tbl_info = self.db.model.users_info

            self.checkLength(tbl.c.identifier, identifier)
            self.checkLength(tbl_info.c.attr_type, attr_type)
            self.checkLength(tbl_info.c.attr_data, attr_data)

            # try to find the user
            q = sa.select(
                tbl_info.c.uid,
            ).where(tbl_info.c.attr_type == attr_type, tbl_info.c.attr_data == attr_data)
            rows = conn.execute(q).fetchall()

            if rows:
                return rows[0].uid

            if _race_hook is not None:
                _race_hook(conn)

            # try to do both of these inserts in a transaction, so that both
            # the new user and the corresponding attributes appear at the same
            # time from the perspective of other masters.
            transaction = conn.begin_nested()
            inserted_user = False
            try:
                r = conn.execute(tbl.insert(), {"identifier": identifier})
                uid = r.inserted_primary_key[0]
                inserted_user = True

                conn.execute(
                    tbl_info.insert(), {"uid": uid, "attr_type": attr_type, "attr_data": attr_data}
                )

                transaction.commit()
            except (sa.exc.IntegrityError, sa.exc.ProgrammingError):
                transaction.rollback()

                # try it all over again, in case there was an overlapping,
                # identical call to findUserByAttr.  If the identifier
                # collided, we'll try again indefinitely; otherwise, only once.
                if no_recurse:
                    raise

                # if we failed to insert the user, then it's because the
                # identifier wasn't unique
                if not inserted_user:
                    identifier = identifiers.incrementIdentifier(256, identifier)
                else:
                    no_recurse = True

                return thd(conn, no_recurse=no_recurse, identifier=identifier)

            conn.commit()
            return uid

        return self.db.pool.do(thd)

    @base.cached("usdicts")
    def getUser(self, uid: int) -> defer.Deferred[UserModel | None]:
        def thd(conn) -> UserModel | None:
            tbl = self.db.model.users
            tbl_info = self.db.model.users_info

            q = tbl.select().where(tbl.c.uid == uid)
            users_row = conn.execute(q).fetchone()

            if not users_row:
                return None

            # gather all attr_type and attr_data entries from users_info table
            q = tbl_info.select().where(tbl_info.c.uid == uid)
            rows = conn.execute(q).fetchall()

            return self._model_from_row(users_row, rows)

        return self.db.pool.do(thd)

    def _model_from_row(self, users_row, attribute_rows=None):
        attributes = None
        if attribute_rows is not None:
            attributes = {row.attr_type: row.attr_data for row in attribute_rows}
        return UserModel(
            uid=users_row.uid,
            identifier=users_row.identifier,
            bb_username=users_row.bb_username,
            bb_password=users_row.bb_password,
            attributes=attributes,
        )

    # returns a Deferred that returns a value
    def getUserByUsername(self, username: str | None) -> defer.Deferred[UserModel | None]:
        def thd(conn) -> UserModel | None:
            tbl = self.db.model.users
            tbl_info = self.db.model.users_info

            q = tbl.select().where(tbl.c.bb_username == username)
            users_row = conn.execute(q).fetchone()

            if not users_row:
                return None

            # gather all attr_type and attr_data entries from users_info table
            q = tbl_info.select().where(tbl_info.c.uid == users_row.uid)
            rows = conn.execute(q).fetchall()

            return self._model_from_row(users_row, rows)

        return self.db.pool.do(thd)

    def getUsers(self) -> defer.Deferred[list[UserModel]]:
        def thd(conn) -> list[UserModel]:
            tbl = self.db.model.users
            rows = conn.execute(tbl.select()).fetchall()

            return [self._model_from_row(row, attribute_rows=None) for row in rows]

        return self.db.pool.do(thd)

    # returns a Deferred that returns None
    def updateUser(
        self,
        uid: int | None = None,
        identifier: str | None = None,
        bb_username: str | None = None,
        bb_password: str | None = None,
        attr_type: str | None = None,
        attr_data: str | None = None,
        _race_hook=None,
    ):
        def thd(conn):
            tbl = self.db.model.users
            tbl_info = self.db.model.users_info
            update_dict = {}

            # first, add the identifier is it exists
            if identifier is not None:
                self.checkLength(tbl.c.identifier, identifier)
                update_dict['identifier'] = identifier

            # then, add the creds if they exist
            if bb_username is not None:
                assert bb_password is not None
                self.checkLength(tbl.c.bb_username, bb_username)
                self.checkLength(tbl.c.bb_password, bb_password)
                update_dict['bb_username'] = bb_username
                update_dict['bb_password'] = bb_password

            # update the users table if it needs to be updated
            if update_dict:
                q = tbl.update().where(tbl.c.uid == uid)
                conn.execute(q, update_dict)

            # then, update the attributes, carefully handling the potential
            # update-or-insert race condition.
            if attr_type is not None:
                assert attr_data is not None

                self.checkLength(tbl_info.c.attr_type, attr_type)
                self.checkLength(tbl_info.c.attr_data, attr_data)

                try:
                    self.db.upsert(
                        conn,
                        tbl_info,
                        where_values=(
                            (tbl_info.c.uid, uid),
                            (tbl_info.c.attr_type, attr_type),
                        ),
                        update_values=((tbl_info.c.attr_data, attr_data),),
                        _race_hook=_race_hook,
                    )
                    conn.commit()
                except (sa.exc.IntegrityError, sa.exc.ProgrammingError):
                    # someone else beat us to the punch inserting this row;
                    # let them win.
                    conn.rollback()

        return self.db.pool.do_with_transaction(thd)

    # returns a Deferred that returns None
    def removeUser(self, uid):
        def thd(conn):
            # delete from dependent tables first, followed by 'users'
            for tbl in [
                self.db.model.change_users,
                self.db.model.users_info,
                self.db.model.users,
            ]:
                conn.execute(tbl.delete().where(tbl.c.uid == uid))

        return self.db.pool.do_with_transaction(thd)

    # returns a Deferred that returns a value
    def identifierToUid(self, identifier) -> defer.Deferred[int | None]:
        def thd(conn) -> int | None:
            tbl = self.db.model.users

            q = tbl.select().where(tbl.c.identifier == identifier)
            row = conn.execute(q).fetchone()
            if not row:
                return None

            return row.uid

        return self.db.pool.do(thd)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Portions Copyright Buildbot Team Members
# Portions Copyright Steve 'Ashcrow' Milner <smilner+buildbot@redhat.com>
"""
Steps specific to the rpm format.
"""

from buildbot.steps.package.rpm.mock import MockBuildSRPM
from buildbot.steps.package.rpm.mock import MockRebuild
from buildbot.steps.package.rpm.rpmbuild import RpmBuild
from buildbot.steps.package.rpm.rpmlint import RpmLint

__all__ = ['MockBuildSRPM', 'MockRebuild', 'RpmBuild', 'RpmLint']



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from unittest import mock

from twisted.internet import defer
from twisted.internet import reactor
from twisted.spread import pb
from twisted.trial import unittest

from buildbot.clients import usersclient


class TestUsersClient(unittest.TestCase):
    def setUp(self):
        # patch out some PB components and make up some mocks
        self.patch(pb, 'PBClientFactory', self._fake_PBClientFactory)
        self.patch(reactor, 'connectTCP', self._fake_connectTCP)

        self.factory = mock.Mock(name='PBClientFactory')
        self.factory.login = self._fake_login
        self.factory.login_d = defer.Deferred()

        self.remote = mock.Mock(name='PB Remote')
        self.remote.callRemote = self._fake_callRemote
        self.remote.broker.transport.loseConnection = self._fake_loseConnection

        # results
        self.conn_host = self.conn_port = None
        self.lostConnection = False

    def _fake_PBClientFactory(self):
        return self.factory

    def _fake_login(self, creds):
        return self.factory.login_d

    def _fake_connectTCP(self, host, port, factory):
        self.conn_host = host
        self.conn_port = port
        self.assertIdentical(factory, self.factory)
        self.factory.login_d.callback(self.remote)

    def _fake_callRemote(self, method, op, bb_username, bb_password, ids, info):
        self.assertEqual(method, 'commandline')
        self.called_with = {
            "op": op,
            "bb_username": bb_username,
            "bb_password": bb_password,
            "ids": ids,
            "info": info,
        }
        return defer.succeed(None)

    def _fake_loseConnection(self):
        self.lostConnection = True

    def assertProcess(self, host, port, called_with):
        self.assertEqual(
            [host, port, called_with], [self.conn_host, self.conn_port, self.called_with]
        )

    @defer.inlineCallbacks
    def test_usersclient_info(self):
        uc = usersclient.UsersClient('localhost', "user", "userpw", 1234)
        yield uc.send(
            'update', 'bb_user', 'hashed_bb_pass', None, [{'identifier': 'x', 'svn': 'x'}]
        )

        self.assertProcess(
            'localhost',
            1234,
            {
                "op": 'update',
                "bb_username": 'bb_user',
                "bb_password": 'hashed_bb_pass',
                "ids": None,
                "info": [{"identifier": 'x', "svn": 'x'}],
            },
        )

    @defer.inlineCallbacks
    def test_usersclient_ids(self):
        uc = usersclient.UsersClient('localhost', "user", "userpw", 1234)
        yield uc.send('remove', None, None, ['x'], None)

        self.assertProcess(
            'localhost',
            1234,
            {"op": 'remove', "bb_username": None, "bb_password": None, "ids": ['x'], "info": None},
        )
